#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2011, GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '$Revision$'

import operator

import gc3libs
from gc3libs.utils import progressive_number


# since the code for `Id` comparison methods is basically the same for
# all methods, we use a decorator-based approach to reduce boilerplate
# code...  Oh, how do I long for LISP macros! :-)
def _Id_make_comparison_function(op):
    """
    Return a function that compares two `Id` objects with the
    passed relational operator. Discards the function being
    decorated.
    """
    def decorate(fn):
        def cmp_fn(self, other):
            try:
                return op((str(self._prefix), self._seqno),
                          (str(other._prefix), other._seqno))
            except AttributeError:
                # fall back to safe comparison as `str`
                gc3libs.log.debug("Wrong job ID: comparing '%s' (%s) with '%s' (%s)"
                                  % (self, type(self), other, type(other)))
                return op(str(self), str(other))
        return cmp_fn
    return decorate


class Id(str):
    """
    An automatically-generated "unique identifier" (a string-like object).
    The unique object identifier has the form "PREFIX.NNN"
    where "NNN" is a decimal number, and "PREFIX" defaults to the
    object class name but can be overridden in the `Id`
    constructor.

    Two object IDs can be compared iff they have the same prefix; in
    which case, the result of the comparison is the same as comparing
    the two sequence numbers.
    """

    def __new__(cls, prefix, seqno):
        """
        Construct a new "unique identifier" instance (a string).
        """
        instance = str.__new__(cls, "%s.%d" % (prefix, seqno))
        instance._seqno = seqno
        instance._prefix = prefix
        return instance

    def __getnewargs__(self):
        return (self._prefix, self._seqno)

    # Rich comparison operators, to ensure `Id` is sorted by numerical value
    @_Id_make_comparison_function(operator.gt)
    def __gt__(self, other):
        pass

    @_Id_make_comparison_function(operator.ge)
    def __ge__(self, other):
        pass

    @_Id_make_comparison_function(operator.eq)
    def __eq__(self, other):
        pass

    @_Id_make_comparison_function(operator.ne)
    def __ne__(self, other):
        pass

    @_Id_make_comparison_function(operator.le)
    def __le__(self, other):
        pass

    @_Id_make_comparison_function(operator.lt)
    def __lt__(self, other):
        pass


class IdFactory(object):
    """
    Automatically generate a "unique identifier" (of class `Id`).
    Object identifiers are temporally unique: no identifier will
    (ever) be re-used, even in different invocations of the program.
    """
    def __init__(self, prefix=None, next_id_fn=None, id_class=Id):
        """
        Construct an `IdFactory` instance whose `new` method returns
        objects of class `id_class` (default: `Id`:class:) with the
        given `prefix` string and whose identifier number is computed
        by an invocation of function `next_id_fn`.

        Function `next_id_fn` must conform to the calling syntax and
        behavior of the `gc3libs.utils.progressive_number`:func:
        (which is the one used by default).
        """
        self._prefix = prefix
        if next_id_fn is None:
            self._next_id_fn = progressive_number
        else:
            self._next_id_fn = next_id_fn
        self._idclass = id_class

    def reserve(self, n):
        """
        Pre-allocate `n` IDs.  Successive invocations of the `Id`
        constructor will return one of the pre-allocated, with a
        potential speed gain if many `Id` objects are constructed in a
        loop.
        """
        assert n > 0, "Argument `n` must be a positive integer"
        IdFactory._seqno_pool.extend(self._next_id_fn(n))
    _seqno_pool = []

    def new(self, obj):
        """
        Return a new "unique identifier" instance (a string).
        """
        if self._prefix is None:
            prefix = obj.__class__.__name__
        else:
            prefix = self._prefix
        if len(IdFactory._seqno_pool) > 0:
            seqno = IdFactory._seqno_pool.pop()
        else:
            seqno = self._next_id_fn()
        return self._idclass(prefix, seqno)


class JobIdFactory(IdFactory):
    """
    Override :class:`IdFactory` behavior and generate IDs starting with a
    lowercase ``job`` prefix.
    """
    def __init__(self, next_id_fn=None):
        IdFactory.__init__(self, 'job', next_id_fn)


## main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="id",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
"""
Unit tests for the OpenStack backend
"""
# Copyright (C) 2012-2013, GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '$Revision$'

# stdlib imports
import os
import tempfile

# 3rd party imports
from nose.tools import assert_equal, assert_false, assert_true, raises, set_trace

# local imports
import gc3libs.config
from gc3libs.backends.openstack import OpenStackLrms
import gc3libs.exceptions

def _setup_config_file(confstr):
    (fd, name) = tempfile.mkstemp()
    f = os.fdopen(fd, 'w+')
    f.write(confstr)
    f.close()
    return name

def test_openstack_variables_are_optionals():
    tmpfile = _setup_config_file("""

[auth/gc3user_ssh]
type = ssh
username = gc3-user

[auth/hobbes]
type=openstack

# Mandatory fields that are set via environment variable
# os_username=USERNAME
# os_password=PASSWORD
# os_project_name=TENANT

[resource/hobbes]
# Mandatory fields that are set via environment variable
# os_auth_url=http://cloud.gc3.uzh.ch:5000/v2.0
os_region=RegionOne
public_key=~/.ssh/id_dsa.pub
name=hobbes
enabled = yes
type=openstack+shellcmd
auth=hobbes
max_cores_per_job = 8
max_memory_per_core = 2
max_walltime = 8
max_cores = 32
architecture = x86_64
keypair_name=keypair
vm_auth=gc3user_ssh
instance_type=m1.small
image_id=a9e98055-5aa3-4636-8fc5-f3b2b4ea66bb
security_group_name=gc3pie_ssh
security_group_rules=tcp:22:22:0.0.0.0/0, icmp:-1:-1:0.0.0.0/0
""")
    cfgvalues = ['username', 'password', 'tenant_name', 'auth_url']
    for name in cfgvalues:
        os.environ['OS_' + name.upper()] =  name

    try:
        cfg = gc3libs.config.Configuration(tmpfile)
        resources = cfg.make_resources()
        assert 'hobbes' in resources

        for name in cfgvalues:
            assert hasattr(resources['hobbes'], 'os_'+name)
            assert_equal(getattr(resources['hobbes'], 'os_'+name), name)
    finally:
        os.remove(tmpfile)


class TestOpenStackLrms(object):
    pass

## main: run tests

if "__main__" == __name__:
    import nose
    nose.runmodule()



#! /usr/bin/env python
"""
Authentication support for the GC3Libs.
"""
# Copyright (C) 2009-2012 GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '2.2.1 version (SVN $Revision: 3977 $)'

import sys

import gc3libs.exceptions

class Auth(object):
    """
    A mish-mash of authorization functions.

    This class actually serves the purposes of:

    - a registry of authorization 'types', mapping internally-assigned
      names to Python classes;
    - storage for the configuration information (which can be
      arbitrary, but should probably be read off a configuration
      file);
    - a factory, returning a 'SomeAuth' object through which clients
      can deal with actual authorization issues (like checking if the
      authorization credentials are valid and getting/renewing them).
    - a cache, that tries to avoid expensive re-initializations of
      `Auth` objects by allowing only one live instance per type, and
      returning it when requested.

    .. admonition:: FIXME

      There are several problems with this approach:

      - the configuration is assumed *static* and cannot be changed after
        the `Auth` instance is constructed.
      - there is no communication between the client class and the
        `Auth` classes.
      - there is no control over the lifetime of the cache; at a
        minimum, it should be settable per-auth-type.
      - I'm unsure whether the mapping of 'type names' (as in the
        `type=...` keyword in the config file) to Python classes
        belongs in a generic factory method or in the configuration
        file reader.  (Probably the former, so the code here would
        actually be right.)
      - The whole `auto_enable` stuff really belongs to the user-interface
        part, which is also hard-coded in the auth classes, and should not be.
    """
    types = {}
    def __init__(self, config, auto_enable):
        self.auto_enable = auto_enable
        self.__auths = { }
        self._config = config
        self._ctors = { }
        for auth_name, auth_params in self._config.iteritems():
            self._ctors[auth_name] = Auth.types[auth_params['type']]


    def add_params(self, **params):
        """
        Add the specified keyword arguments as initialization
        parameters to all the configured auth classes.

        Parameters that have already been specified are silently
        overwritten.
        """
        for auth_name, auth_params in self._config.iteritems():
            auth_params.update(params)


    def get(self, auth_name, **kwargs):
        """
        Return an instance of the `Auth` class corresponding to the
        given `auth_name`, or raise an exception if instanciating the
        same class has given an unrecoverable exception in past calls.

        Additional keyword arguments are passed unchanged to the class
        constructor and can override values specified at configuration time.

        Instances are remembered for the lifetime of the program; if
        an instance of the given class is already present in the
        cache, that one is returned; otherwise, an instance is
        contructed with the given parameters.

        .. caution::

          The `params` keyword arguments are only used if a new
          instance is constructed and are silently ignored if the
          cached instance is returned.

        """
        if auth_name not in self.__auths:
            try:
                params = self._config[auth_name].copy()
                params.update(kwargs)
                a =  self._ctors[auth_name](**dict(params))
            except (AssertionError, AttributeError), ex:
                a = gc3libs.exceptions.ConfigurationError(
                    "Missing required configuration parameters"
                    " in auth section '%s': %s" % (auth_name, str(ex)))
        else:
            a = self.__auths[auth_name]

        if isinstance(a, Exception):
            if isinstance(a, gc3libs.exceptions.UnrecoverableError):
                self.__auths[auth_name] = a
            raise a

        if not a.check():
            if self.auto_enable:
                try:
                    a.enable()
                except gc3libs.exceptions.RecoverableAuthError, x:
                    raise
                except gc3libs.exceptions.UnrecoverableAuthError, x:
                    gc3libs.log.debug("Got exception while enabling auth '%s',"
                                      " will remember for next invocations:"
                                      " %s: %s" % (auth_name, x.__class__.__name__, x))
                    a = x
            else:
                a = gc3libs.exceptions.UnrecoverableAuthError(
                    "No valid credentials of type '%s'"
                    " and `auto_enable` not set." % auth_name)

        self.__auths[auth_name] = a
        return a


    @staticmethod
    def register(auth_type, ctor):
        Auth.types[auth_type] = ctor


class NoneAuth(object):
    """Auth proxy to use when no auth is needed."""
    def __init__(self, **auth):
        try:
            # test validity
            assert auth['type'] == 'none', (
                "Configuration error. Unknown type: %s. Valid type: none"
                % auth.type)
        except AssertionError, x:
            raise gc3libs.exceptions.ConfigurationError(
                'Erroneous configuration parameter: %s' % str(x))

    def is_valid(self):
        return True

    def check(self):
        gc3libs.log.debug("Checking auth: NoneAuth (always successful).")
        return True

    def enable(self):
        return True

Auth.register('none', NoneAuth)
# register additional auth types
# FIXME: it would be nice to have some kind of auto-discovery instead
import gc3libs.authentication.grid
import gc3libs.authentication.ssh
import gc3libs.authentication.ec2
import gc3libs.authentication.openstack

## main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="__init__",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
#   Copyright (C) 2010 2013 GC3, University of Zurich
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script which runs multiple GEOTop simulations over 
a presegmented geographical area (usually 18 segments). The 
workflow is as follows:

- identify the segment in the global map,
- run GEOTop on the sub-parts of the segment,
- agregate the results into a set of output files,

It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gtsub_control --help`` for program usage instructions.

"""
__docformat__ = 'reStructuredText'
__version__ = '$Revision$'


import os
import os.path
import shlex
import time
import sys
import tempfile
import shutil

import tarfile

import gc3libs
import gc3libs.exceptions
from gc3libs.utils import fgrep
from gc3libs import Application, Run, Task
from gc3libs.quantity import Memory, GB
from gc3libs.cmdline import SessionBasedScript, existing_directory, existing_file
from gc3libs.workflow import TaskCollection
from gc3libs.workflow import ParallelTaskCollection, SequentialTaskCollection
from pkg_resources import Requirement, resource_filename 

GEOTOP_INPUT_ARCHIVE = "input.tgz"

class GTSubControlApplication(Application):

    # This part starts the different Applications for the
    # specified simulations boxes.

    def __init__(self, sim_box, inputs, **extra_args):


        # source the execution script
        run_gtsub_control = resource_filename(Requirement.parse("gc3pie"), "gc3libs/etc/run_gtsub_control.sh")
        os.chmod(run_gtsub_control,0777) 
        inputs[run_gtsub_control] = 'run_gtsub_control.sh'

        outputs = {
            './sim/result/':             'sim/result',
            './topoApp_complete.r.Rout': 'topoApp_complete.r.Rout',
            './src/TopoAPP/parfile.r':   'parfile.r',
        }

        Application.__init__(self,
                arguments = [ "./run_gtsub_control.sh", sim_box ],
                inputs = inputs,
                outputs = outputs,
                stdout = 'gt_sub.log',
                join=True,
                **extra_args)


class GTSubControlScript(SessionBasedScript):
    """
    GTSubControl Script is used for submiting the GeoTop Application
    on parametric basics. It takes as arguments the [root] location of the
    all the input files and the [nseq] containing the sequence of elements
    corrisponding to a single simulation box
    """
    version = '1.0'

    def setup_options(self):

        self.add_param("-p", "--parameter-file", dest="par_file",
                        type=existing_file,
                        default=None,
                        help="Indicate the parameter file to be used")

        # change default for the "-m"/"--memory-per-core" option
        self.actions['memory_per_core'].default = 7*GB

    def setup_args(self):
        self.add_param("root", type=existing_directory, help='The root directory where to script is looking for input data')
        self.add_param("nseq", type=str, help='The sequence number of sim boxes to be executed. Formats: INT:INT | INT,INT,INT,..,INT | INT')

    def parse_args(self):
        "Split the nseq argument and set sim_boxes parameters."
        try:
            if self.params.nseq.count(':') == 1:
                start, end = self.params.nseq.split(':')
                self.sim_boxes = range(int(start), int(end)+1)
            elif self.params.nseq.count(',') >= 1:
                simulations = self.params.nseq.split(',');
                self.sim_boxes = [ int(s) for s in simulations ]
            else:
                self.sim_boxes = [ int(self.params.nseq) ];
        except ValueError:
            raise gc3libs.exceptions.InvalidUsage(
                "Invalid argument '%s', use on of the following formats: INT:INT | INT,INT,INT,..,INT | INT " % (nseq,))

    def _prepare_tar(self, simulation_dir):
        """Prepare the input tarball."""
        try:
            gc3libs.log.debug("Compressing input folder in'%s' ", simulation_dir)
            cwd = os.getcwd()
            os.chdir(simulation_dir)
            # check if input archive already present. If yes delete and reuse it
            if os.path.isfile(GEOTOP_INPUT_ARCHIVE):
                try:
                    os.remove(GEOTOP_INPUT_ARCHIVE)
                except OSError, x:
                    gc3libs.log.error("Failed removing '%s': %s: %s",
                                      GEOTOP_INPUT_ARCHIVE, x.__class__, x.message)
                    pass
            tar = tarfile.open(GEOTOP_INPUT_ARCHIVE, "w:gz", dereference=True)
            tar.add('./src')
            tar.add('./sim/_master')
            tar.close()
            os.chdir(cwd)
            yield (tar.name, GEOTOP_INPUT_ARCHIVE)
        except Exception, x:
             gc3libs.log.error("Failed creating input archive '%s': %s: %s",
                                os.path.join(simulation_dir, GEOTOP_INPUT_ARCHIVE),
                                x.__class__,x.message)
             raise

    def _get_output_dir_naming(self, paramfile):
        """
        Get the strings needed for the naming of the output directory.
        """
        # parse the parfile and take "Nclust", "col", "beg"
        # and "end" values needed for output directory naming.

        output_dir_naming_strings = []

        # Get the nclust value from the parfile.r
        lines = list(gc3libs.utils.fgrep("Nclust=", paramfile))
        if len(lines) != 1:
          raise RuntimeError(
              "Expecting one and only one `Nclust=` line in %s, found %d"
              % (paramfile, len(lines)))
        line = lines[0].strip() 
        start, end = line.split('=')
        if (end == "" or start == ""):
            raise ValueError("Nclust option in the parfile.r is not properly set."
                             " It is set like %s, must be Nclust=Value." % ( line ))
        nclust = "Nc"+end
        output_dir_naming_strings.append(nclust)

        # Get the col value from the parfile.r
        lines = list(gc3libs.utils.fgrep("col=", paramfile))
        if len(lines) != 1:
          raise RuntimeError(
              "Expecting one and only one `col=` line in %s, found %d"
              % (paramfile, len(lines)))
        line = lines[0].strip()
        start, end = line.split('=')
        if (end == "" or start == ""):
            raise ValueError("col option in the parfile.r is not properly set."
                            " It is set like %s, must be col=Value." % ( line ))
        col = end
        col = col.replace('"','')
        output_dir_naming_strings.append(col)

        # Get the beginning value out of the "beg" line
        lines = list(gc3libs.utils.fgrep("beg <-", paramfile))
        if len(lines) != 1:
          raise RuntimeError(
              "Expecting one and only one `beg=` line in %s, found %d"
              % (paramfile, len(lines)))
        line = lines[0].strip() 
        line_elements = line.split(' ')
        beg = line_elements[2].replace('"','')
        beg = beg.replace('/','')
        output_dir_naming_strings.append(beg)

        # Get the ending value out of the "end" line
        lines = list(gc3libs.utils.fgrep("end <-", paramfile))
        if len(lines) != 1:
          raise RuntimeError(
              "Expecting one and only one `end=` line in %s, found %d"
              % (paramfile, len(lines)))
        line = lines[0].strip()       
        line_elements = line.split(' ')
        end = line_elements[2].replace('"','')
        end = end.replace('/','')
        output_dir_naming_strings.append(end)

        return output_dir_naming_strings

    def new_tasks(self, extra):

        # Call the _prepare_tar method which recreates
        # the tarball each time.
        inputs = dict(self._prepare_tar(self.params.root))

        # Manage the parfile.r and call the _get_output_dir_formatting
        # method which gives back the strings needed for the output dir
        # naming.
        paramfile = self.params.root + '/src/TopoAPP/parfile.r'
        if self.params.par_file:
                inputs[self.params.par_file] = 'parfile.r'
                paramfile = self.params.par_file

        out_dir_nm = self._get_output_dir_naming(paramfile)

        # create tasks for the specified sumulation boxes
        for sim_box in self.sim_boxes:
            extra_args = extra.copy()
            ## specifiedcify the jobname by root and number of box seq
            jobname = ('GTSC_nS%s_%s_%s_%s_%s' 
                        % (sim_box, out_dir_nm[0], out_dir_nm[1], 
                            out_dir_nm[2], out_dir_nm[3])) 
            extra_args['jobname'] = jobname
            yield GTSubControlApplication(
                sim_box, # pass the simulation box number
                inputs.copy(),
                **extra_args)

## main: run tests

if "__main__" == __name__:
        import gtsub_control
        gtsub_control.GTSubControlScript().run()



    #! /usr/bin/env python
#

# Copyright (C) 2011 University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__version__ = '$Revision$'
__author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>, Benjamin Jonen <benjamin.jonen@bf.uzh.ch>'
# summary of user-visible changes
__changelog__ = """

"""
__docformat__ = 'reStructuredText'

import gc3libs.debug
import numpy as np
import re, os, sys
import logbook
from threading import Lock

class wrapLogger():
    def __init__(self, loggerName = 'myLogger', streamVerb = 'DEBUG', logFile = 'logFile'):
        self.loggerName = loggerName
        self.streamVerb = streamVerb
        self.logFile    = logFile
        logger = getLogger(loggerName = self.loggerName, streamVerb = self.streamVerb, logFile = self.logFile)
        self.wrappedLog = logger

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['wrappedLog']
        return state
    def __setstate__(self, state):
        self.__dict__ = state
        logger = getLogger(loggerName = self.loggerName, streamVerb = self.streamVerb, logFile = self.logFile)
        self.wrappedLog = logger
        
    def __getattr__(self, attr):
        # see if this object has attr
        # NOTE do not use hasattr, it goes into
        # infinite recurrsion
        if attr in self.__dict__:
            # this object has it
            return getattr(self, attr)
        # proxy to the wrapped object
        return getattr(self.wrappedLog, attr)
    
    def __hasattr__(self, attr):
        if attr in self.__dict__:
            return getattr(self, attr)
        return getattr(self.wrappedLog, attr)
    


def getLogger(loggerName = 'mylogger.log', streamVerb = 'DEBUG', logFile = 'log'):

    # Get a logger instance.
    logger = logbook.Logger(name = loggerName)
    
    # set up logger
    mySH = logbook.StreamHandler(stream = sys.stdout, level = streamVerb.upper(), format_string = '{record.message}', bubble = True)
    mySH.format_string = '{record.message}'
    logger.handlers.append(mySH)
    if logFile:
        myFH = logbook.FileHandler(filename = logFile, level = 'DEBUG', bubble = True)
        myFH.format_string = '{record.message}' 
        logger.handlers.append(myFH)   
    
    try:
        stdErr = list(logbook.handlers.Handler.stack_manager.iter_context_objects())[0]
        stdErr.pop_application()
    except: 
        pass
    return logger


def lower(npStrAr):
    return np.fromiter((x.lower() for x in npStrAr.flat),
                       npStrAr.dtype, npStrAr.size)


def flatten(lst):
    for elem in lst:
        if type(elem) in (tuple, list):
            for i in flatten(elem):
                yield i
        else:
            yield elem


def str2tuple(strIn, conv=int):
    """
    Convert tuple-like strings to real tuples.
    Example::

      >>> str2tuple('(1,2,3,4)')
      (1, 2, 3, 4)

    Enclosing parentheses can be omitted::

      >>> str2tuple('1,2,3')
      (1, 2, 3)

    An optional conversion function can be specified as second argument
    `conv` (defaults to `int`)::

      >>> str2tuple('1,2,3', float)
      (1.0, 2.0, 3.0)
    """
    strIn = strIn.strip()
    if strIn.startswith('('):
        strIn = strIn[1:-1] # removes the leading and trailing brackets
    items = strIn.split(',')
    return tuple(conv(x) for x in items)


class getIndex():
# Python bug http://stackoverflow.com/questions/1152238/python-iterators-how-to-dynamically-assign-self-next-within-a-new-style-class
# Cannot use new type class, that is: class getIndex(object) does not allow setting next method on the fly. 
    """
    Iterator that yields loop indices for an arbitrary nested loop.
    Inputs: base - array of elements in each dimension
    Output: loopIndex - array of current loop index
    """    

    def __init__(self, base, restr = None, direction = 'rowWise'):
        '''
          Direction specifies which direction is incremented. For matrix A = [1 , 2; rowwise means (0,0), (0,1)
          while columnwise means (0,0), (1, 0).                               3 , 4]
        '''
        if isinstance(base, int):
            baseList = []
            baseList.append(base)
            base = baseList
        self.base = np.array(base)
        self.loopIndex = np.zeros(len(self.base), dtype=np.int16)
        self.iteration = 1
        self.direction = direction

        if isinstance(restr, np.ndarray):
            self.restr = restr[0].lower()
        elif restr:      
            self.restr = restr.lower()
        else:
            self.restr = restr
        diagnolStrings = ['diagnol', 'diag']
        if self.restr in diagnolStrings:
 #           getIndex.next = self.nextDiag
            self.next = self.nextDiag
        else:
#            getIndex.next = self.nextStd
            self.next = self.nextStd

        if direction.lower() == 'rowwise':
            self.increment = self.incrementRowWise
        elif direction.lower() == 'columnwise':
            self.increment = self.incrementColumnWise
        else:
            print('cannot understand direction')
            sys.exit()

    def nextDiag(self):
        if self.iteration > 1:
            self.loopIndex += 1
        self.iteration += 1
        if all(self.loopIndex == min(self.base)):
            raise StopIteration
        return self.loopIndex.tolist()

    def __iter__(self):
        return self
    #getIndex(self.base, self.restr, self.direction)

    def incrementRowWise(self):
        for ix in xrange(len(self.base) - 1, -1, -1):
            if self.loopIndex[ix] == self.base[ix] - 1:
                self.loopIndex[ix] = 0
            else:
                self.loopIndex[ix] += 1
                return

    def incrementColumnWise(self):
        for ix in xrange(0, len(self.base), +1):
            if self.loopIndex[ix] == self.base[ix] - 1:
                self.loopIndex[ix] = 0
            else:
                self.loopIndex[ix] += 1
                return

    def lowerTr(self):
        for ix in xrange(0, len(self.base) - 1) :
            if self.loopIndex[ix] < self.loopIndex[ix + 1]:
                return False
        return True

    def diagnol(self):
        isDiagnolIndex = all([x == self.loopIndex[0] for x in self.loopIndex])
        # Skip if not diagnol
        return not isDiagnolIndex  

    def nextStd(self):
        skip = True
        while skip == True:
            if self.iteration > 1:
                self.increment()

            if list(self.loopIndex) == [0] * len(self.base) and self.iteration > 1:
                raise StopIteration

            self.iteration += 1

            if self.restr == 'lowertr':
                skip = self.lowerTr()
            elif self.restr == None or self.restr == 'none':
                skip = False
            else:
                raise gc3libs.exceptions.InvalidArgument(
                    "Unknown restriction '%s'" % self.restr)

        return self.loopIndex.tolist()


@gc3libs.debug.trace
def extractVal(ixVar, vals, index):
    """
    Return variable value to be updated. 
    Functions allows to specify more involved retrieval. 
    """
    return str2vals(vals[ixVar])[index[ixVar]]


@gc3libs.debug.trace
def str2vals(strIn):
    """
    strIn: string containing different vals
    out:   np.array of different vals
    Function can be used to store val vectors in one overall vector and then unpack the string. 
    """
    if 'linspace' in strIn:
        out = np.array([])
        while strIn:
            if re.match('\s*linspace', strIn):
                (linSpacePart, strIn) = re.match('(\s*linspace\(.*?\)\s*)[,\s*]*(.*)', strIn).groups()
                args = re.match('linspace\(([(0-9\.\s-]+),([0-9\.\s-]+),([0-9\.\s-]+)\)', linSpacePart).groups()
                args = [ float(arg) for arg in args] # assume we always want float for linspace
                linSpaceVec = np.linspace(args[0], args[1], args[2])
                out = np.append(out, linSpaceVec)
            elif re.match('\s*[0-9\.]*\s*,', strIn):
                (valPart, strIn) = re.match('(\s*[0-9\.]*\s*)[,\s*]*(.*)', strIn).groups()
                valPart = valPart.strip()
                if '.' in valPart:
                    valPart = float(valPart)
                else:
                    valPart = int(valPart)
                out = np.append(out, valPart)
        return out
    else:
        return str2mat(strIn)

def format_newVal(newVal):
    import re
    if re.match("^[-]*[0-9]*\.[0-9e-]*$", str(newVal)):
#    if '.' in str(newVal):
        newValMat = ('%25.15f' % float(newVal)).strip()
    else:
        try:
            # try to convert to integer, and use decimal repr
            newValMat = str(int(newVal))
        except ValueError: 
            # then it's a string
            newValMat = newVal
    return newValMat



@gc3libs.debug.trace
def update_parameter_in_file(path, varIn, paraIndex, newVal, regexIn):
    _loop_regexps = {
        'bar-separated':(r'([a-z]+[\s\|]+)'
                         r'(\w+)' # variable name
                         r'(\s*[\|]+\s*)' # bars and spaces
                         r'([\w\s\.,;\[\]\-]+)' # value
                         r'(\s*)'),
        'space-separated':(r'(\s*)'
                           r'(\w+)' # variable name
                           r'(\s+)' # spaces (filler)
                           r'([\w\s\.,;\[\]\-]+)' # values
                           r'(\s*)'), # spaces (filler)
    }
    isfound = False
    if regexIn in _loop_regexps.keys():
        regexIn = _loop_regexps[regexIn]
    paraFileIn = open(path, 'r')
    paraFileOut = open(path + '.tmp', 'w')
    for line in paraFileIn:
        #print "Read line '%s' " % line
        if not line.rstrip(): continue
        (a, var, b, oldValMat, c) = re.match(regexIn, line.rstrip()).groups()
        gc3libs.log.debug("Read variable '%s' with value '%s' ...", var, oldValMat)
        if var == varIn:
            isfound = True
            oldValMat = str2mat(oldValMat)
            if oldValMat.shape == (1,):
                newValMat = newVal
            else:
                newValMat = oldValMat
                newValMat[paraIndex] = newVal
                newValMat = mat2str(newValMat)
            gc3libs.log.debug("Will change variable '%s' to value '%s' ...", var, newValMat)
        else:
            newValMat = oldValMat
        paraFileOut.write(a + var + b + newValMat + c + '\n')
    paraFileOut.close()
    paraFileIn.close()
    # move new modified content over the old
    os.rename(path + '.tmp', path)
    if not isfound:
        gc3libs.log.critical('update_parameter_in_file could not find parameter in sepcified file')


def safe_eval(s):
    """
    Try to parse `s` as an integer; if successful, return its integer
    value.  Else try to parse it as a float, and return that value.
    Otherwise, return `s` converted to a string.

    Example::

      >>> safe_eval("42")
      42
      >>> safe_eval("42.0")
      42.0
      >>> safe_eval("forty-two")
      'forty-two'

    """
    # must attempt conversion to `int` before we try `float`,
    # since every `int` literal is also a valid `float` literal
    try:
        return int(s)
    except ValueError:
        pass
    try:
        return float(s)
    except ValueError:
        pass
    return str(s)


@gc3libs.debug.trace
def str2mat(strIn):
    """
    strIn: String containing matrix
    out:   np.array with translated string elements. 
    """
    if (',' not in strIn) and (';' not in strIn):
        # single value
        return np.array([ safe_eval(strIn.strip()) ])
    elif ';' not in strIn: # vector
        # remove wrapping '[' and ']' (if any)
        try:
            start = strIn.index('[')
            end = strIn.rindex(']')
            strIn = strIn[start+1:end]
        except ValueError: # no '[' or ']'
            pass
        # a vector is a ','-separated list of values
        return np.array([ safe_eval(element.strip())
                          for element in strIn.split(',') ])
    else: # matrix
        mat = [ ]
        # remove wrapping '[' and ']' (if any)
        try:
            start = strIn.index('[')
            end = strIn.rindex(']')
            strIn = strIn[start+1:end]
        except ValueError: # no '[' or ']'
            pass
        # a matrix is a ';'-separated list of rows;
        # a row is a ','-separated list of values
        return np.array([ [ safe_eval(val) for val in row.split(',') ]
                          for row in strIn.split(';') ])


def mat2str(matIn, fmt='%.2f '):
    """
    Return a string representation of 2D array `matIn`.
    A matrix is printed as a `;`-separated list of rows,
    where each row is a `,`-separated list of values.
    All enclosed in `[` and `]`.
    """
    # admittedly, this is not that clear...
    return ('[ ' +
            str.join(";", [
                str.join(",",
                         [(fmt % val) for val in row])
                for row in matIn ])
            + ' ]')


def getParameter(fileIn, varIn, regexIn = '(\s*)([a-zA-Z0-9]+)(\s+)([a-zA-Z0-9\.\s,;\[\]\-]+)(\s*)'):
    import re
    _loop_regexps = {
        'bar-separated':(r'([a-z]+[\s\|]+)'
                         r'(\w+)' # variable name
                         r'(\s*[\|]+\s*)' # bars and spaces
                         r'([\w\s\.,;\[\]\-]+)' # value
                         r'(\s*)'),
        'space-separated':(r'(\s*)'
                           r'(\w+)' # variable name
                           r'(\s+)' # spaces (filler)
                           r'([\w\s\.,;\[\]\-]+)' # values
                           r'(\s*)'), # spaces (filler)
    }
    if regexIn in _loop_regexps.keys():
        regexIn = _loop_regexps[regexIn]
    #  print('updateParameter inputs: \n --- \n {0} \n {1} \n {2} \n {3} \n {4} \n ---'.format(fileIn, varIn, paraIndex, newVal, regexIn))
    paraFile = open(fileIn)
    lines = paraFile.readlines()
    paraFile.close()
    for ixLine, line in enumerate(lines):
        (a, var, b, paraVal, c) = re.match(regexIn, line.rstrip()).groups()
    #      print('var=', var)
        if var == varIn:
            return paraVal
    print('variable {} not in parameter file {}'.format(varIn, fileIn))

def walklevel(some_dir, level=1):
    '''
    from http://stackoverflow.com/questions/229186/os-walk-without-digging-into-directories-below
    '''
    some_dir = some_dir.rstrip(os.path.sep)
    assert os.path.isdir(some_dir)
    num_sep = some_dir.count(os.path.sep)
    for root, dirs, files in os.walk(some_dir):
        yield root, dirs, files
        num_sep_this = root.count(os.path.sep)
        if num_sep + level <= num_sep_this:
            del dirs[:]
            
def emptyFun():
    return None 


if __name__ == '__main__':   
    x = getIndex([3,3], 'lowerTr')
    for i in x:
        print(i)


#! /usr/bin/env python
#
#   gcodeml.py -- Front-end script for submitting multiple CODEML jobs to SMSCG.
#
#   Copyright (C) 2010-2012 GC3, University of Zurich
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple CODEML jobs to SMSCG.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gcodeml --help`` for program usage instructions.
"""
__version__ = '2.2.1 version (SVN $Revision$)'
# summary of user-visible changes
__changelog__ = """
  2013-02-18:
    * Output files places in same folder as inputs.
  2012-12-05:
    * Allow requesting a specific version of CODEML/PAML
      via a command-line option.
  2012-06-20:
    * Save extended job information if the store allows it
      (presently, only `SqlStore` does).  The definition of
      the extra information table is hard-coded in this script,
      in the `GCodemlScript._make_session` method.
  2011-07-22:
    * Re-submit a CODEML job if it exits with nonzero exit code.
    * If no option ``-x``/``--codeml-executable`` is given on the
      command-line, `gcodeml` will now search for the
      ``APPS/BIO/CODEML-4.4.3`` application tag and use the remotely
      provided executable.
  2011-03-22:
    * New option ``-x`` to set the path to the ``codeml`` executable.
  2011-03-21:
    * Changed argument processing logic: no longer submit each
      ``.ctl`` file as a separate jobs, instead assume all the
      files in a single directory are related and bundle them into
      a single job.
    * New option ``-O`` to upload output files to a GridFTP
      server, instead of downloading them to a local destination.
  2011-02-08:
    * Initial release, forked off the ``ggamess`` sources.
"""
__author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>'
__docformat__ = 'reStructuredText'



# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: http://code.google.com/p/gc3pie/issues/detail?id=95
if __name__ == '__main__':
    import gcodeml
    gcodeml.GCodemlScript().run()


# std module imports
import os
import os.path
import re
import sys

import sqlalchemy as sqla

# gc3 library imports
import gc3libs
import gc3libs.exceptions
import gc3libs.persistence.sql
import gc3libs.session
import gc3libs.utils

from gc3libs.application.codeml import CodemlApplication
from gc3libs.cmdline import SessionBasedScript, executable_file
from gc3libs.persistence.accessors import GET, GetValue
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask


## retry policy

class CodemlRetryPolicy(RetryableTask):

    def retry(self):
        # return True or False depending whether the application
        # should be re-submitted or not.
        # The actual CodemlApplication is available as `self.task`,
        # so for instance `self.task.valid[0]` is `True` iff the
        # H0.mlc file is present and processed correctly.
        # gc3libs.log.debug("CodemlRetryPolicy called!")
        # for now, do the default (see: gc3libs/__init__.py)
        to_retry = RetryableTask.retry(self)
        gc3libs.log.debug("CodemlRetryPolicy called with retry [%s]" % str(to_retry))
        return to_retry


## the script itself

class GCodemlScript(SessionBasedScript):
    """
Scan the specified INPUTDIR directories recursively for '.ctl' files,
and submit a CODEML job for each input file found; job progress is
monitored and, when a job is done, its '.mlc' file is retrieved back
into the same directory where the '.ctl' file is (this can be
overridden with the '-o' option).

The `gcodeml` command keeps a record of jobs (submitted, executed and
pending) in a session file (set name with the '-s' option); at each
invocation of the command, the status of all recorded jobs is updated,
output from finished jobs is collected, and a summary table of all
known jobs is printed.  New jobs are added to the session if new input
files are added to the command line.

Options can specify a maximum number of jobs that should be in
'SUBMITTED' or 'RUNNING' state; `gcodeml` will delay submission
of newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            # `CodemlRetryPolicy` is the top-level object here,
            # so only print information about it.
#            stats_only_for = CodemlRetryPolicy,
             stats_only_for = gc3libs.application.codeml.CodemlApplication,
            )


    def setup_options(self):
        # add some options to the std ones
        self.add_param("-O", "--output-url", action="store",
                       dest="output_base_url", default="",
                       help="Upload output files to this URL,"
                       "which must be a protocol that ARC supports."
                       "(e.g., 'gsiftp://...')")
        self.add_param("-R", "--verno", metavar="VERNO",
                       dest="verno", default=CodemlApplication.DEFAULT_CODEML_VERSION,
                       help="Request the specified version of CODEML/PAML"
                       " (default: %(default)s).")
        self.add_param("-x", "--codeml-executable", metavar="PATH",
                       action="store", dest="codeml",
                       type=executable_file, default=None,
                       help=("Local path to the CODEML executable."
                             " By default, request the CODEML-%s run time tag"
                             " and use the remotely-provided application."
                           % CodemlApplication.DEFAULT_CODEML_VERSION))
        # change default for the "-o"/"--output" option
        self.actions['output'].default = 'NAME'


    def new_tasks(self, extra):
        """Implement the argument -> jobs mapping."""
        ## process additional options
        if self.params.codeml is not None:
            if not os.path.isabs(self.params.codeml):
                self.params.codeml = os.path.abspath(self.params.codeml)
            gc3libs.utils.test_file(self.params.codeml, os.R_OK|os.X_OK)

        ## collect input directories/files
        def contain_ctl_files(paths):
            for path in paths:
                if path.endswith('.ctl'):
                    return True
            return False

        input_dirs = set()
        total_n_ctl_files = 0

        for path in self.params.args:
            self.log.debug("Now processing input argument '%s' ..." % path)
            if not os.path.isdir(path):
                raise gc3libs.exceptions.InvalidUsage(
                    "Argument '%s' is not a directory path." % path)

            # recursively scan for input files
            for dirpath, dirnames, filenames in os.walk(path):
                if contain_ctl_files(filenames):
                    input_dirs.add(os.path.realpath(dirpath))

        self.log.debug("Gathered input directories: '%s'"
                       % str.join("', '", input_dirs))

        for dirpath in input_dirs:
            # gather control files; other input files are automatically
            # pulled in by CodemlApplication by parsing the '.ctl'
            ctl_files = [ os.path.join(dirpath, filename)
                          for filename in os.listdir(dirpath)
                          if filename.endswith('.ctl') ]

            total_n_ctl_files += len(ctl_files) # AK: DEBUG
            # check if the dir contains exactly two control files (*.H0.ctl and *.H1.ctl)
            if len(ctl_files) != 2:
                raise gc3libs.exceptions.InvalidUsage(
                    "The input directory '%s' must contain exactly two control files." % dirpath)

            self.log.debug("Gathered control files: '%s':" % str.join("', '", ctl_files))

            kwargs = extra.copy()

            ## create new CODEML application instance

            # Python 2.4 does not allow named arguments after a
            # variable-length positional argument list (*args), so we
            # need to pass the named arguments as part of the `extra_args`
            # dictionary.
            kwargs['codeml'] = self.params.codeml
            kwargs['version'] = self.params.verno
            kwargs['requested_memory'] = self.params.memory_per_core
            kwargs['requested_cores'] = self.params.ncores
            kwargs['requested_walltime'] = self.params.walltime

            jobname = (os.path.basename(dirpath) or dirpath) + '.out'

            if self.params.output == 'NAME':
                # This is the 'default' setting
                # gcodeml overrides it
                # Results in same folder as inputs
                kwargs['output_dir'] = os.path.join(dirpath,'.compute')
                kwargs['result_dir'] = dirpath
            else:
                # command line option takes precendence
                # in this case, just set output_dir to what has been specified
                # as '--output' option
                kwargs['output_dir'] = self.make_directory_path(self.params.output, jobname)

            # FIXME: should this reflect the same policy as 'output_dir' ?
            # set optional arguments (path to 'codeml' binary, output URL, etc.)
            if self.params.output_base_url != "":
               kwargs['output_base_url'] = self.params.output_base_url

            app = CodemlApplication(*ctl_files, **kwargs)

            # yield new job
            yield (
                jobname, # unique string tagging this job; duplicate jobs are discarded
                CodemlRetryPolicy, # the task constructor
                [ # the following parameters are passed to the
                    # `CodemlRetryPolicy` constructor:
                    app,     # = task; the codeml application object defined above
                    3        # = max_retries; max no. of retries of the task
                ],
                # extra keyword parameters to `CodemlRetryPolicy`
                dict()
                )

        self.log.debug("Total number of control files: %d", total_n_ctl_files) # AK: DEBUG

    def _make_session(self, session_uri, store_url):
        return gc3libs.session.Session(
            session_uri,
            store_url,
            extra_fields = {
                # NB: enlarge window to at least 150 columns to read this table properly!
                sqla.Column('class',              sqla.TEXT())    : (lambda obj: obj.__class__.__name__)                                              , # task class
                sqla.Column('name',               sqla.TEXT())    : GetValue()             .jobname                                                   , # job name
                sqla.Column('executable',         sqla.TEXT())    : GetValue(default=None) .arguments[0]                        ,#.ONLY(CodemlApplication), # program executable
                sqla.Column('output_path',        sqla.TEXT())    : GetValue(default=None) .output_dir                        ,#.ONLY(CodemlApplication), # fullpath to codeml output directory
                sqla.Column('input_path',         sqla.TEXT())    : _get_input_path                                                                   , # fullpath to codeml input directory
                sqla.Column('mlc_exists_h0',      sqla.TEXT())    : GetValue(default=None) .exists[0]                         ,#.ONLY(CodemlApplication), # exists codeml *.H0.mlc output file
                sqla.Column('mlc_exists_h1',      sqla.TEXT())    : GetValue(default=None) .exists[1]                         ,#.ONLY(CodemlApplication), # exists codeml *.H1.mlc output file
                sqla.Column('mlc_valid_h0',       sqla.TEXT())    : GetValue(default=None) .valid[0]                          ,#.ONLY(CodemlApplication), #.attrid codeml *.H0.mlc output file
                sqla.Column('mlc_valid_h1',       sqla.TEXT())    : GetValue(default=None) .valid[1]                          ,#.ONLY(CodemlApplication), #.attrid codeml *.H1.mlc output file
                sqla.Column('cluster',            sqla.TEXT())    : GetValue(default=None) .execution.resource_name           ,#.ONLY(CodemlApplication), # cluster/compute element
                sqla.Column('worker',             sqla.TEXT())    : GetValue(default=None) .hostname                          ,#.ONLY(CodemlApplication), # hostname of the worker node
                sqla.Column('cpu',                sqla.TEXT())    : GetValue(default=None) .cpuinfo                           ,#.ONLY(CodemlApplication), # CPU model of the worker node
                sqla.Column('codeml_walltime_h0', sqla.INTEGER()) : GetValue()             .time_used[0]                      ,#.ONLY(CodemlApplication), # time used by the codeml H0 run (sec)
                sqla.Column('codeml_walltime_h1', sqla.INTEGER()) : GetValue()             .time_used[1]                      ,#.ONLY(CodemlApplication), # time used by the codeml H1 run (sec)
                sqla.Column('aln_len',            sqla.TEXT())    : GetValue()             .aln_info['aln_len']                                    , # alignement length
                sqla.Column('seq',                sqla.TEXT())    : GetValue()             .aln_info['n_seq']                                      , # num of sequences
                sqla.Column('requested_walltime', sqla.INTEGER()) : _get_requested_walltime_or_none                           , # requested walltime, in hours
                sqla.Column('requested_cores',    sqla.INTEGER()) : GetValue(default=None) .requested_cores                   ,#.ONLY(CodemlApplication), # num of cores requested
                sqla.Column('tags',               sqla.TEXT())    : GetValue()             .tags[0]                           ,#.ONLY(CodemlApplication), # run-time env.s (RTE) requested; e.g. 'APPS/BIO/CODEML-4.4.3'
                sqla.Column('used_walltime',      sqla.INTEGER()) : GetValue(default=None) .execution.used_walltime           ,#.ONLY(CodemlApplication), # used walltime
                sqla.Column('lrms_jobid',         sqla.TEXT())    : GetValue(default=None) .execution.lrms_jobid              ,#.ONLY(CodemlApplication), # arc job ID
                sqla.Column('original_exitcode',  sqla.INTEGER()) : GetValue(default=None) .execution.original_exitcode       ,#.ONLY(CodemlApplication), # original exitcode
                sqla.Column('used_cputime',       sqla.INTEGER()) : GetValue(default=None) .execution.used_cputime            ,#.ONLY(CodemlApplication), # used cputime in sec
                # returncode = exitcode*256 + signal
                sqla.Column('returncode',         sqla.INTEGER()) : GetValue(default=None) .execution.returncode              ,#.ONLY(CodemlApplication), # returncode attr
                sqla.Column('queue',              sqla.TEXT())    : GetValue(default=None) .execution.queue                   ,#.ONLY(CodemlApplication), # exec queue _name_
                sqla.Column('time_submitted',     sqla.FLOAT())   : GetValue(default=None) .execution.timestamp['SUBMITTED']  ,#.ONLY(CodemlApplication), # client-side submission (float) time
                sqla.Column('time_terminated',    sqla.FLOAT())   : GetValue(default=None) .execution.timestamp['TERMINATED'] ,#.ONLY(CodemlApplication), # client-side termination (float) time
                sqla.Column('time_stopped',       sqla.FLOAT())   : GetValue(default=None) .execution.timestamp['STOPPED']    ,#.ONLY(CodemlApplication), # client-side stop (float) time
                sqla.Column('retried',            sqla.INTEGER()) : GetValue(default=None) .retried                           .ONLY(CodemlRetryPolicy), # num of times job has been retried
                })

# auxiliary getter functions, used in `GCodemlScript_make_session` above
def _get_input_path_CodemlApplication(job):
    # we know that each CodemlApplication has only two .ctl
    # files as input, and they belong in the same directory,
    # so just pick the first one
    for src_url in job.inputs:
        if src_url.path.endswith('.ctl'):
            return os.path.dirname(src_url.path)

def _get_input_path(job):
    if isinstance(job, gc3libs.application.codeml.CodemlApplication):
        return _get_input_path_CodemlApplication(job)
    elif isinstance(job, CodemlRetryPolicy):
        return _get_input_path_CodemlApplication(job.task)
    else:
        return None

def _get_requested_walltime_or_none(job):
    if isinstance(job, gc3libs.application.codeml.CodemlApplication):
        return job.requested_walltime.amount(hours)
    else:
        return None



#! /usr/bin/env python
#
#   sepjobs.py -- Classify GAMESS output files according to keyword matches
#
#   Copyright (C) 2010, 2011 GC3, University of Zurich
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Classify GAMESS output files according to keyword matches.
"""
__version__ = '2.2.1 version (SVN $Revision$)'
# summary of user-visible changes
__changelog__ = """
  2011-08-18:
    * update of config.file now after file inspection and job assignment
    * worked up movement of unassigned jobs to "none-found"
  2011-08-16:
    * added profiling
  2011-08-15:
    * witing new lines to config.file now seems to work (although I don't know why)
    * now moving folder of each job instead of moving files if each inspected file resides in its own folder, autom. determing common path part of file paths
    * added profiling
  2011-08-09:
    * moving files into subfolders for key strings now works
    * automatic distinct. between new and old search by subfolder "none_found" now works
    * added working dir. and source dir. to handle new and old search
    * still problem with writing to config.file
  2011-08-08:
    * removed support for several input dir.s
    * default search rootpath now is current path of script
    * opening of file in loop for search item now conditioned by search range of item
    * search in entire file now outside loop for search items
    * worked up parse_kwfile
    * still problem with file moving
  2011-08-04:
    * Removed csv-stuff, cannot deal with ""I/O", ERROR"
    * Now using ":" to separate searchstring from rest and "," for residual fields
    * Found problem with list comprehension and os.walk in "search_for_input_directories"
    * Added "done"-flag to make script ingoring already processed lines in config. file
  2011-08-03:
    * Use `.csv` file for configuration.
  2011-07-27:
    * Initial release.
"""
__author__ = 'Timm Reumann <timm.reumann@uzh.ch>, Riccardo Murri <riccardo.murri@gmail.com>'
__docformat__ = 'reStructuredText'


import os
import re
import sys
import fnmatch
import shutil
import cli.log


from gc3libs.cmdline import existing_file, existing_directory


sys.setcheckinterval(1000)    # interpret. check interval for threads and signals

def CLEANLIST(itmlst, delstr) :
        delpos = []; p = 0                                      # list of deleting position, position
        for itm in itmlst :
                if re.match(delstr, itm) :              # "\" to suppress interpret. of "." by python
                        delpos.append(p)
                p += 1
        ndp = 0                                                         # no. of deleted positions
        for p in delpos :                       
                p -= ndp                                # convert. pos. in old dirlist to pos. in current one
                itmlst.pop(p)
#               del dirlist(p)                  # gives error "can't delete function call"
                ndp += 1
        return itmlst


def search_for_input_directories(root, name):
    """
    Return list of directories under `root` (at any nesting level)
    whose with user specified input dir. `name`.
    """
    name = name.strip('/')
    print 'DEBUG: stripped name of input folder:', name
    result = []
    for rootpath, dirs, files in os.walk(root):
        matching_dirs = []
        # removing hidden directories from directory list
        """
        if os.path.basename(rootpath).startswith(r'.'):
            print "DEBUG: current path:\n", rootpath
        if rootpath == root: 
            print "DEBUG: rootpath:\n", rootpath
            print "DEBUG: raw dirs:\n", dirs
        
        dirs = [ d for d in dirs if not d.startswith(r'.') ]
        This produces the correct output according to print, but os.walk
        somehow ignores it - still goes into .dirname
        """
        dirs = CLEANLIST(dirs, '\.')
#        if rootpath == root: 
#            print "DEBUG: cleaned dirs:\n", dirs
        for dirname in dirs:
#            if dirname in names:  # support of several input folders undesirable
            if dirname == name:
                result.append(os.path.join(rootpath, dirname))
                matching_dirs.append(dirname)
        # removing assigned directories from directory list
        for dirname in matching_dirs:
            dirs.remove(dirname)
#    print "DEBUG: dirs matching", name, ":\n", result    
    return result


def search_for_none_found(root):
    old_search = False
    if 'none_found' in os.listdir(root):
        old_search = True
        root = os.path.join(root, 'none_found')
#    print 'DEBUG: dirlist for search of none_found:\n', os.listdir(root)
    return root, old_search


def search_for_input_files(root, ext):
    """
    Return list of files under `dir` (or any if its subdirectories)
    that match extension `ext`.
    """
    result = [ ]
    for rootpath, dirs, files in os.walk(root):
        # removing hidden directories from directory list
#        dirs = [ d for d in dirs if not d.startswith('.') ]  see above
        dirs = CLEANLIST(dirs, '\.')
        for filename in files:
            if filename.endswith(ext):
                result.append(os.path.join(rootpath, filename))
    return result


class StringAdaptor(str):
    """
    Augment a `str` object with `search` and `match` methods with
    interface compatible with that exposed by the `re` objects.
    """
    def search(self, container):
        return (self in container)
    def match(self, container):
        return container.startswith(self)


def parse_kwfile(filename, old_search):
    """
    Read `filename` and return a dictionary mapping a folder name into
    a corresponding regular expression to search for.
    """
    number = re.compile(r'[0-9]+\.?[0-9]*')
    done = re.compile(r'done[ ,]*')
    folder_exists = False
 #   configfile = open(filename, 'r')
    result = [] ; newlns = []
    with open(filename, 'r') as configfile:
        for line in configfile: 
            newlns.append(line)
            if line == '\n' or line.startswith('#') or line.isspace():
#                print "DEBUG: ignoring line %r"% line
                continue                        # skip empty or comment lines
            if done.search(line.rsplit(':', 1)[1]):
                if old_search:
#                    print "DEBUG: ignoring line %r"% line
                    continue            # skip lines already used for separation
            else:
                newlns[len(newlns)-1] = newlns[len(newlns)-1].rstrip(' \n') + ', done\n'
#            print "DEBUG: processing line %r"% line
#            print "DEBUG: new lines %r"% newlns[len(newlns)-1]
            what_to_search, rest = line.rsplit(':', 1)
#            print "DEBUG: search string:", str(what_to_search)
            start = 0.0 ; end = 1.0
            rest = rest.replace(' ', '').rstrip('\n') + ',,,'
            field = rest.split(',')
            searchtype = field[0]
            foldername = field[1]
            if number.match(field[2]):
                start = float(field[2]) / 100.0
            if number.match(field[3]):
                end = float(field[3]) / 100.0
            print "DEBUG: search string, searchtype, foldername, start, end:\n", '"', what_to_search, '"', searchtype, foldername, start, end
            if searchtype in ('regexp', 're'):
                result.append([foldername, re.compile(what_to_search), start, end, 
                               folder_exists])
            else:                                   # literal string search
                result.append([foldername, re.compile(what_to_search), start, end, 
                               folder_exists])
        configfile.close()
    return result, newlns


## main function

@cli.log.LoggingApp
def sepjobs(cmdline):
    # `cmdline` is pyCLI's interface object
    cmdline.log.debug("Starting.")
    
    # look for directory name under the search root (if given)
    dirs = search_for_input_directories(cmdline.params.search_root,
                                            cmdline.params.inpdir)
    wrkdir = dirs[0]
    cmdline.log.debug("Working directory: %s", wrkdir)
    srcdir, old_search = search_for_none_found(wrkdir)
    none_found_exists = old_search
    # look for input files and common path of all files
    files = search_for_input_files(srcdir, cmdline.params.file)
#    cmdline.log.debug("List of input files: %s", files)
        # conditioning by cmdln argument "-m" + number later
#    if cmdline.params.dirlvl == -1:
    commonpath = os.path.commonprefix(files).rsplit('/',1)[0]
#    else if cmdline.params.dirlvl == 0:
        
#    else:
#        commonpath = os.path.commonprefix(files).rsplit('/',cmdline.params.dirlvl)[0]
    cmdline.log.debug("common prefix input file paths:\n %s", commonpath)
    # parse classification file
    srchkey, newlns = parse_kwfile(cmdline.params.kwfile, old_search)
    cmdline.log.debug("Searching keywords: %s", srchkey)
    # create indices
    index = { }
    for i in xrange(0, len(srchkey)):
        index[srchkey[i][0]] = set()
    cmdline.log.debug("index: %s", index)
    index['none_found'] = set(files)
    
    def act_on_file(filepath, foldername, assigned, folder_exists, old_search): 
        """Actions to perform when a file matches a classification keyword."""
        if assigned:
            index[foldername].add(filepath) 
            index['none_found'].remove(filepath)
        """
        dirpth0 = os.path.dirname(filepath)
        print dirpth0.replace(commonpath, 
            os.path.join(wrkdir, foldername), 1)
        """
        if cmdline.params.move:
#              cmdline.log.info("Going to move file")
            if not folder_exists: 
                os.mkdir(os.path.join(wrkdir, foldername))
                folder_exists = True
#              cmdline.log.info("Moving file")
            dirpath0 = os.path.dirname(filepath)
            if dirpath0 != commonpath:
                dirpath1 = dirpath0.replace(commonpath, 
                           os.path.join(wrkdir, foldername), 1)
                try:
                    os.makedirs(dirpath1)
                except OSError:
                    print '\r'
                os.rename(dirpath0, dirpath1)
            else:
                os.rename(filepath, os.path.join(wrkdir, foldername, 
                    os.path.basename(filepath)))
        return assigned, folder_exists    
    
    # let's rock!
    failrt = 0
    hitrt = 0
    Nsrchkeys = len(srchkey)
    for filepath in files:
#        cmdline.log.info("Now processing file '%s' ...", filepath)
        inputfile = open(filepath, 'r')
        start0 = 1.0 ; end0 = 0.0
        assigned = False
        for i in xrange(0, Nsrchkeys):
#            cmdline.log.info("Now looking for '%s' ...", srchkey[i][4])
            if srchkey[i][2] < start0 or srchkey[i][3] > end0: 
                # compute range
                flsize = os.path.getsize(filepath)
                btpos = int(flsize * srchkey[i][2])
                btrng = int(flsize * (srchkey[i][3] - srchkey[i][2])) + 1
                # read content
                inputfile.seek(btpos, 0)         # going to byte position in file
                content = inputfile.read(btrng)  # reading byte range from byte pos.
                start0 = srchkey[i][2]
                end0 = srchkey[i][3]
            match = srchkey[i][1].search(content)       
            if match:
#                cmdline.log.info("Found match, folder_exists='%s'", srchkey[i][4])
                hitrt += 1
                assigned, srchkey[i][4] = act_on_file(filepath, srchkey[i][0], True, 
                                                      srchkey[i][4], old_search)
                break
        # retry search with whole file
        if not assigned:               
            content = inputfile.read()
            for i in xrange(0, Nsrchkeys):
                match = srchkey[i][1].search(content)
                if match:
                    failrt += 1
                    assigned, srchkey[i][4] = act_on_file(filepath, srchkey[i][0], True, 
                                                          srchkey[i][4], old_search)
                    # how to avoid copying regexp back to srchkey ?  
                    break
        if not assigned:
            assigned, none_found_exists = act_on_file(filepath, 'none_found', False, 
                                                      none_found_exists, old_search)
        inputfile.close()
    with open(cmdline.params.kwfile, 'w') as configfile:
        print "DEBUG: writing new lines to config.file"
        configfile.writelines(newlns)
        configfile.close()
    # print summary
    for foldername, filenames in index.iteritems():
        print "== %s ==" % foldername, "no. of files:", len(index[foldername])
#        for filename in sorted(filenames):
#            print ("    " + filename)
    cmdline.log.info("failure rate = %d, hit rate = %d", failrt, hitrt)


## command-line parameters

sepjobs.add_param('kwfile', type=existing_file,
                  help="Path to the file containing foldername to string search mappings.")
sepjobs.add_param('inpdir', type=existing_directory, # no support of several input dir.s
                  help="Directory where files to analyze are located.")
sepjobs.add_param('-f', '--file', '--fl',
                  dest='file', metavar='EXT', default='.out',
                  help="Restrict search to file with this extension."
                  " (default: %(default)s)")
# sepjobs.add_param('-m', '--move',
#                   dest='dirlvl', default=-1,
#                   help="Move files with their folders accord. to classification. Optionally with directory level (0 for pure file moving, 3 for moving file with folders up to 3rd level")
sepjobs.add_param('-m', '--move',
                  dest='move', action='store_true', default=False,
                  help="Move files into folders named after their classification keyword.")
                  
sepjobs.add_param('-S', '--search-root', metavar='DIR',
                  dest='search_root',
                  type=existing_directory, default=os.getcwd(), 
                  help="Search for input directories under the directory tree rooted at DIR. Must be a COMPLETE PATH e.g. ~/Desktop/Project"
                  "  (Default: '%(default)s')")

import cProfile
import profile
import pstats
if __name__ == '__main__':
    sepjobs.run()
    """
    # determining overhead of cProfiler
    pr = profile.Profile()
    cumovrhd = 0.0
    for i in range(5):
        ovrhd = pr.calibrate(10000)
        print 'overhead', i, '=', ovrhd 
        cumovrhd += ovrhd
    print 'average overhead =', cumovrhd / 5.0
    profile.Profile.bias = cumovrhd / 5.0
    cProfile.run('sepjobs.run()', 'sepjobs.prfl')
    p = pstats.Stats('sepjobs.prfl')
    p.sort_stats('time', 'calls').print_stats(10)
    """
       
""" 
config file for sepjobs.py:

# sample classifier file for sepjobs.py
#
# Each line has this form:
#    
#   searchstring: literal|regexp, foldername, [start, end], ['done']
#
               
gracefully: regexp, gracefully, 90.0, , done         

#I/O ERROR: literal, io_error, 80,, done, ,      
#ddikick.x: Timed out: literal, timed_out,, , , ,
#Failed creating: literal, failed_creat, 90,,,,,,,
#ILLEGAL GENERAL  BASIS FUNCTION REQUESTED: literal, ILLEGAL_BASISFUNC, 80,    
#ILLEGAL EXTENDED BASIS FUNCTION REQUESTED.: literal, ILLEGAL_EXTBASISFUNC, 80  , 
#JACOBI DIAGONALIZATION FAILS: literal, JACOBI_DIAG_FAILS, 80  , done
#ILLEGAL          BASIS FUNCTION REQUESTED: literal, ILLEGAL_____BASISFUNC, 80  
"""



#! /usr/bin/env python
#
#   gbiointeract.py -- Front-end script for submitting multiple
#   `biointeract` jobs.
#
#   Copyright (C) 2013 GC3, University of Zurich
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple `biointeract` jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gbiointeract --help`` for program usage instructions.
"""

__version__ = '2.2.1 version (SVN $Revision$)'
__author__ = 'Antonio Messina <antonio.s.messina@gmail.com>'
__docformat__ = 'reStructuredText'

import decimal
import itertools
import os
import tempfile

import gc3libs
from gc3libs import Application, Run
from gc3libs.cmdline import SessionBasedScript
from gc3libs.workflow import ChunkedParameterSweep
from gc3libs.utils import write_contents

def arange(start, stop=None, step=1, precision=None):
    """arange generates a set of Decimal values over the
    range [start, stop) with step size step

    arange([start,] stop, [step [,precision]])

    Courtesy of Nisan Haramati:
    http://code.activestate.com/recipes/66472-frange-a-range-function-with-float-increments/#c14
    """

    if stop is None:
        for x in xrange(int(ceil(start))):
            yield x
    else:
        # find precision
        if precision is not None:
            decimal.getcontext().prec = precision
        # convert values to decimals
        start = decimal.Decimal(start)
        stop = decimal.Decimal(stop)
        step = decimal.Decimal(step)
        # create a generator expression for the index values
        indices = (
            i for i in xrange(
                0,
                ((stop-start)/step).to_integral_value()
            )
        )
        # yield results
        for i in indices:
            yield float(start + step*i)


class GBiointeractApplication(Application):
    """
    Class to wrap execution of a single `biointeract` program.
    """

    application_name = "gbiointeract"

    def __init__(self,
                 executable,
                 cell_diffusion,
                 public_good_durability,
                 public_good_diffusion,
                 death_rate,
                 make_movie,
                 num_cells=None,
                 num_trials=None,
                 **extra_args):
        extra_args.setdefault('requested_cores', 1)

        self.executable = executable
        self.cell_diffusion = cell_diffusion
        self.public_good_diffusion = public_good_diffusion
        self.public_good_durability = public_good_durability
        self.death_rate = death_rate

        arguments = ['./'+os.path.basename(executable),
                     '-c', str(float(cell_diffusion)),
                     '-p', str(float(public_good_diffusion)),
                     '-d', str(int(public_good_durability)),
                     '-x', str(float(death_rate)),
                     '-D', 'data']
        if make_movie:
            arguments.append('-m')
        if num_cells:
            arguments.extend(['--num-cells', num_cells])
        if num_trials:
            arguments.extend(['--num-trials', num_trials])
            
        extra_args['jobname'] = "GBiointeract_cdiff:" + \
        "%f_pgdiff:%f_dur:%f_deathrate:%f" % (cell_diffusion,
                                              public_good_diffusion,
                                              public_good_durability, death_rate)
        if 'output_dir' in extra_args:
            extra_args['output_dir'] = os.path.join(
                os.path.dirname(extra_args['output_dir']),
                extra_args['jobname'])

        executable_script = """#!/bin/sh

# Create data directory
mkdir data

# Run the script

%s

# Compress all output files
gzip data/*.txt
""" % str.join(' ', arguments)
        try:
            (fd, self.tmp_filename) = tempfile.mkstemp(prefix='gc3pie-gbiointeract_')
            write_contents(self.tmp_filename, executable_script)
            os.chmod(self.tmp_filename, 0755)
        except Exception, ex:
            gc3libs.log.debug("Error creating execution script."
                              "Error type: %s. Message: %s" % (type(ex), ex.message))
            raise

        data_files = [
            'data/allele_nonproducer.txt.gz',
            'data/allele_producer.txt.gz',
            'data/doubling_size_time.txt.gz',
            'data/draw.txt.gz',
            'data/final_numbers.txt.gz',
            'data/final_values.txt.gz',
            'data/initial_values.txt.gz',
            'data/log.txt.gz',
            'data/plot_values.txt.gz',
            'data/public_goods.txt.gz',
            'data/statistics.txt.gz',
            'data/warnings.txt.gz',
            ]

        Application.__init__(self,
                             ['./gbiointeract.sh'],
                             [executable, (self.tmp_filename, 'gbiointeract.sh')],        # inputs
                             data_files + ["gbiointeract.sh"],  # outputs
                             stdout="gbiointeract.out",
                             stderr="gbiointeract.err",
                             **extra_args)

    def terminated(self):
        """
        Remove temporary script file
        """
        try:
            os.remove(self.tmp_filename)
        except Exception, ex:
            gc3libs.log.error("Failed removing temporary file %s. " % self.tmp_filename +
                          "Error type %s. Message %s" % (type(ex), str(ex)))

class GBiointeractTaskCollection(ChunkedParameterSweep):
    def __init__(self,
                 executable,
                 cell_diffusion_range,
                 public_good_diffusion_range,
                 public_good_durability_range,
                 death_rate_range,
                 chunk_size,
                 make_movie,
                 num_cells,
                 num_trials,
                 **extra_args):
        if not chunk_size:
            chunk_size = step
        self.executable = executable
        self.extra_args = extra_args
        self.make_movie = make_movie
        self.num_cells = num_cells
        self.num_trials = num_trials

        self.combinations = list(itertools.product(
                 cell_diffusion_range,
                 public_good_durability_range,
                 public_good_diffusion_range,
                 death_rate_range))
        self.curr = 0

        ChunkedParameterSweep.__init__(self,
                                       1,
                                       len(self.combinations)+1, # it has to be at least 2
                                       1,
                                       chunk_size,
                                       **extra_args)

    def new_task(self, param, **extra):
        (cd, d, pg, dr) = self.combinations[self.curr]
        self.curr += 1

        return GBiointeractApplication(
            self.executable,
            cd, d, pg, dr, self.make_movie,
            self.num_cells, self.num_trials,
            **self.extra_args)


class GBiointeractScript(SessionBasedScript):
    """
    Run multiple instances of `biointeract` with a combination of
    supplied parameters.
    """
    version = __version__
    application_name = 'gbiointeract'

    def _parse_range(self, string):
        """
        Parse a string in the form N[:END:STEP] and returns an iterator.

        Raises a ValueError if the string does not match
        """
        params = string.split(':')
        if len(params) not in (1, 3):
            raise ValueError("Invalid syntax for range `%s'" % string)

        if len(params) == 3:
            return arange(*string.split(':'))
        else:
            return iter((float(params[0]),))

    def setup_options(self):
        self.add_param("-c", "--cell-diffusion",
                       help="In the form N[:END:STEP]. If only `N` is "
                       "supplied, will use only that value, otherwise "
                       "will use all the values in the range from `N` "
                       "to `END` (exclusive) using `STEP` increments")
        self.add_param("-p", "--public-good-diffusion",
                       help="In the form N[:END:STEP]. If only `N` is "
                       "supplied, will use only that value, otherwise "
                       "will use all the values in the range from `N` "
                       "to `END` (exclusive) using `STEP` increments")
        self.add_param("-d", "--public-good-durability",
                       help="In the form N[:END:STEP]. If only `N` is "
                       "supplied, will use only that value, otherwise "
                       "will use all the values in the range from `N` "
                       "to `END` (exclusive) using `STEP` increments")
        self.add_param("-x", "--death-rate",
                       help="In the form N[:END:STEP]. If only `N` is "
                       "supplied, will use only that value, otherwise "
                       "will use all the values in the range from `N` "
                       "to `END` (exclusive) using `STEP` increments")
        self.add_param("-M", "--make-movie", action="store_true",
                       help="Also produce the `draw.txt` file containing "
                       "the cell positions at each iteration.")
        self.add_param("--num-cells", help="Maximum number of cells.")
        self.add_param("--num-trials", help="Number of trials to run.")

    def setup_args(self):
        self.add_param("executable", nargs="?", default='biointeract',
                       metavar='EXECUTABLE',
                       help="Path to the biointeract executable.")

    def parse_args(self):
        if self.params.cell_diffusion:
            self.params.cell_diffusion_range = self._parse_range(
                self.params.cell_diffusion)

        if self.params.public_good_diffusion:
            self.params.public_good_diffusion_range = self._parse_range(
                self.params.public_good_diffusion)

        if self.params.public_good_durability:
            self.params.public_good_durability_range = self._parse_range(
                self.params.public_good_durability)

        if self.params.death_rate:
            self.params.death_rate_range = self._parse_range(
                self.params.death_rate)

        self.params.executable = os.path.abspath(self.params.executable)

    def new_tasks(self, extra):
        # Check that the required parameters have been passed to the script
        if not self.params.cell_diffusion:
            raise gc3libs.exceptions.InvalidUsage("-c, --cell-diffusion argument is required")

        if not self.params.public_good_diffusion:
            raise gc3libs.exceptions.InvalidUsage("-p, --public-good-diffusion argument is required")

        if not self.params.public_good_durability:
            raise gc3libs.exceptions.InvalidUsage("-d, --public-good-durability argument is required")

        if not self.params.death_rate:
            raise gc3libs.exceptions.InvalidUsage("-x, --death-rate argument is required")

        # Check that the `biointeract` executable exists.
        if not os.path.exists(self.params.executable):
            raise gc3libs.exceptions.InvalidUsage("Biointeract binary file `%s` does not exists. Please, re-run specifying a valid path to the biointeract program." % self.params.executable)
        elif not os.path.isfile(self.params.executable):
            raise gc3libs.exceptions.InvalidUsage("Invalid biointeract binary `%s`. Please, re-run specifying a valid path to the biointeract program." % self.params.executable)

        return [GBiointeractTaskCollection(
            self.params.executable,
            self.params.cell_diffusion_range,
            self.params.public_good_diffusion_range,
            self.params.public_good_durability_range,
            self.params.death_rate_range,
            self.params.max_running,
            self.params.make_movie,
            self.params.num_cells,
            self.params.num_trials,
            **extra.copy()
            )]


if __name__ == '__main__':
    import gbiointeract
    gbiointeract.GBiointeractScript().run()



#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2012, GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '$Revision$'

import sys
import logging
import os
import shutil
import tempfile

from nose.tools import assert_equal

from gc3libs import Application, Run, configure_logger, create_engine
from gc3libs.workflow import SequentialTaskCollection

loglevel = logging.ERROR
configure_logger(loglevel, "test_isse_335")

class MySequentialCollection(SequentialTaskCollection):
    def next(self, x):
        """ensure that the next() is called only once per task."""
        if not hasattr(self, 'next_called_n_times'):
            self.next_called_n_times = 1
        else:
            self.next_called_n_times += 1
        return SequentialTaskCollection.next(self, x)

class test_issue_335(object):
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        CONF_FILE="""
[auth/dummy]
type = ssh
username = dummy

[resource/localhost]
enabled = true
auth = dummy
type = shellcmd
frontend = localhost
transport = local
max_cores_per_job = 2
max_memory_per_core = 2
max_walltime = 8
max_cores = 2
architecture = x86_64
override = False
resourcedir = %s
"""
        self.cfgfile = os.path.join(self.tmpdir, 'gc3pie.conf')
        self.resourcedir = os.path.join(self.tmpdir, 'shellcmd.d')
        fp = open(self.cfgfile, 'w')
        fp.write(CONF_FILE % self.resourcedir)
        fp.close()

    def test_issue(self):
        """Test that SequentialTasksCollection goes in terminated state when all of its tasks are in TERMINATED state."""
        self.ptasks = 5
        task = MySequentialCollection(
            [
                Application(
                    ['echo','test1'],
                    [],[],
                    os.path.join(self.tmpdir, 'test.%d.d' % i)) for i in range(self.ptasks)
                ]
            )
        engine = create_engine(self.cfgfile, auto_enable_auth=True)
        engine.add(task)
        while True:
            engine.progress()

            if len([t for t in task.tasks if t.execution.state == Run.State.TERMINATED]) == self.ptasks:
                engine.progress()
                assert_equal(task.execution.state, Run.State.TERMINATED)
                break
        assert_equal(task.next_called_n_times, self.ptasks)

    def tearDown(self):
        shutil.rmtree(self.tmpdir)


    
if "__main__" == __name__:
    import nose
    nose.runmodule()



#! /usr/bin/env python
#
"""
Driver script for running Turbomole basis benchmarks
on the SMSCG infrastructure.
"""
# Copyright (C) 2011-2012 GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__version__ = '$Revision$'
__author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>'
# summary of user-visible changes
__changelog__ = """
  2011-06-20:
    * Copy the ``coord`` file into each generated ``ridft`` and ``ricc2`` directory.
    * Move the ``ricc2`` directory at the leaf of the generated tree.
  2011-05-06:
    * Record RIDFT/RICC2 output into a `ridft.out`/`ricc2.out` file
      in the corresponding `output/` subdirectory.
    * Workaround for Issue 95: now we have complete interoperability
      with GC3Utils.
"""
__docformat__ = 'reStructuredText'


# workaround Issue 95, see: http://code.google.com/p/gc3pie/issues/detail?id=95
if __name__ == "__main__":
    import gricomp
    gricomp.GRICompScript().run()


# stdlib imports
import ConfigParser
import csv
import math
import os
import os.path
import shutil
import sys
import types

# interface to Gc3libs
import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.application.turbomole import TurbomoleApplication, TurbomoleDefineApplication
from gc3libs.cmdline import SessionBasedScript
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import StagedTaskCollection, ParallelTaskCollection
from gc3libs.template import Template, expansions
import gc3libs.utils


## support classes

# list basis set names (as accepted by "define") here;
# ORDER IS IMPORTANT: the `acceptable_*_basis_set` functions
# will only allow a jkbas, cbas or cabs if it's not *earlier*
# than the orbital basis; i.e. if the orbital basis is
# ``aug-cc-pVQZ`` then jkbas cannot be ``aug-cc-pVTZ``.
basis_set_names = [
    'aug-cc-pVTZ',
    'aug-cc-pVQZ',
    'aug-cc-pV5Z',
    ]

def acceptable_ridft_basis_set(extra_args):
    """Define which combination of orbital and JK basis are valid."""
    def order(k): # small aux function
        return basis_set_names.index(extra_args[k])
    orb_basis_nr = order('ORB_BASIS')
    jkbas_basis_nr = order('RIJK_BASIS')
    # only allow a jkbas if it's not *earlier* than the orbital basis;
    # i.e. if the orbital basis is ``aug-cc-pVQZ`` then jkbas cannot
    # be ``aug-cc-pVTZ``.
    if orb_basis_nr > jkbas_basis_nr:
        return False
    # otherwise, the basis combination is acceptable
    return True

def acceptable_ricc2_basis_set(extra_args):
    """Define which combination of CABS and CBAS are valid."""
    def order(k): # small aux function
        return basis_set_names.index(extra_args[k])
    orb_basis_nr = order('ORB_BASIS')
    cabs_basis_nr = order('CABS_BASIS')
    cbas_basis_nr = order('CBAS_BASIS')
    # only allow a cbas or cabs if it's not *earlier*
    # than the orbital basis; i.e. if the orbital basis is
    # ``aug-cc-pVQZ`` then cbas cannot be ``aug-cc-pVTZ``.
    if orb_basis_nr > cbas_basis_nr:
        return False
    if orb_basis_nr > cabs_basis_nr:
        return False
    # otherwise, the basis combination is acceptable
    return True

# TURBOMOLE's ``define`` input for the RIDFT step
RIDFT_DEFINE_IN = """\

${TITLE}
a coord
*
no
b all ${ORB_BASIS}
*
eht



scf
conv
8

ri
on

rijk
on
jkbas
b all ${RIJK_BASIS}
*
m ${RIDFT_MEMORY}

*
"""

# TURBOMOLE's ``define`` input for the RICC2 step
RICC2_DEFINE_IN = """\





cc
freeze
*
cbas
b all ${CBAS_BASIS}
*
memory ${RICC2_MEMORY}
denconv 1.0d-8
ricc2
mp2 energy only
*
f12
pairenergy on
*
cabs
b all ${CABS_BASIS}
*
*
*
"""

def _make_define_in(path, contents):
    """
    Write `contents` to a file named ``define.in`` in directory
    `path`.  Return the full path to the written file.
    """
    define_in_filename = os.path.join(path, 'define.in')
    define_in_file = open(define_in_filename, 'w')
    define_in_file.write(str(contents))
    define_in_file.close()
    return define_in_filename


class LocalApplication(Application):
    """
    Force an Application to run on `localhost`.
    """
    def compatible_resources(self, resources):
        """
        Only `localhost` matches.
        """
        return [ lrms for lrms in resources if lrms._resource.name == 'localhost' ]


class NonLocalApplication(Application):
    """
    Force an Application *not* to run on `localhost`.
    """
    def compatible_resources(self, resources):
        """
        Only `localhost` does not match.
        """
        return [ lrms for lrms in resources if lrms._resource.name != 'localhost' ]


class NonLocalTurbomoleApplication(NonLocalApplication, TurbomoleApplication):
    """
    Like `TurbomoleApplication`, but force it *not* to run on `localhost`.
    """
    pass


class NonLocalTurbomoleDefineApplication(NonLocalApplication, TurbomoleDefineApplication):
    """
    Like `TurbomoleDefineApplication`, but force it *not* to run on `localhost`.
    """
    pass


class XmlLintApplication(LocalApplication):
    # xmllint --schema /links/xml-recources/xml-validation/CML3scheme.xsd ./control_*.xml 1>/dev/null 2>validation.log
    def __init__(self, turbomole_output_dir, output_dir,
                 validation_log='validation.log',
                 schema='/links/xml-recources/xml-validation/CML3scheme.xsd',
                 **extra_args):
        self.validation_log = validation_log
        # find the control_*.xml in the TURBOMOLE output directory
        control_xml = None
        for filename in os.listdir(turbomole_output_dir):
            if filename.startswith('control_') and filename.endswith('.xml'):
                control_xml = os.path.join(turbomole_output_dir, filename)
                break
        if control_xml is None:
            raise ValueError("Cannot find a 'control_*.xml' file in directory '%s'"
                             % turbomole_output_dir)
        gc3libs.Application.__init__(
            self,
            executable='xmllint',
            arguments = [ '--schema', schema, control_xml ],
            inputs = [ control_xml ],
            outputs = [ validation_log ],
            output_dir = output_dir,
            stdout = None,
            stderr = validation_log,
            **extra_args)

    def terminated(self):
        validation_logfile = open(os.path.join(self.output_dir, self.validation_log), 'r')
        validation_log_contents = validation_logfile.read()
        validation_logfile.close()
        if 'validates' in validation_log_contents:
            self.execution.returncode = (0, 0) # SUCCESS
        else:
            self.execution.returncode = (0, 1) # FAIL


class XmlDbApplication(LocalApplication):
    # /opt/eXist/bin/client.sh -u fox -m "/db/home/fox/${projectdir}" -p control_* -P 'tueR!?05' -s 1>/dev/null 2>&1
    def __init__(self, turbomole_output_dir, output_dir, db_dir, db_user, db_pass, db_log='db.log', **extra_args):
        # find the control_*.xml in the TURBOMOLE output directory
        control_xml = None
        for filename in os.listdir(turbomole_output_dir):
            if filename.startswith('control_') and filename.endswith('.xml'):
                control_xml = os.path.join(turbomole_output_dir, filename)
                break
        if control_xml is None:
            raise ValueError("Cannot find a 'control_*.xml' file in directory '%s'"
                             % turbomole_output_dir)
        # pre-process control_*.xml to remove the 'xmlns' part,
        # which confuses eXist
        to_remove = 'xmlns="http://www.xml-cml.org/schema"'
        os.rename(control_xml, control_xml + '.orig')
        control_xml_file_in = open(control_xml + '.orig', 'r')
        control_xml_file_out = open(control_xml, 'w')
        for line in control_xml_file_in:
            line = line.replace(to_remove, '')
            control_xml_file_out.write(line)
        control_xml_file_in.close()
        control_xml_file_out.close()

        gc3libs.Application.__init__(
            self,
            executable='/opt/eXist/bin/client.sh',
            arguments = [
                '-u', db_user,
                '-P', db_pass,
                '-m', db_dir,
                '-p', control_xml,
                '-s',
            ],
            inputs = [ control_xml ],
            outputs = [ db_log ],
            output_dir = output_dir,
            stdout = db_log,
            stderr = db_log,
            join=True,
            **extra_args)


class TurbomoleAndXmlProcessingPass(StagedTaskCollection):
    """
    Run a TURBOMOLE application, then validate the 'control_*.xml'
    file produced, and store it into an eXist database.

    """
    def __init__(self, name, turbomole_application, output_dir,
                 db_dir, db_user, db_pass,
                 **extra_args):
        self.turbomole_application = turbomole_application
        self.output_dir = output_dir
        self.db_dir = db_dir
        self.db_user = db_user
        self.db_pass = db_pass
        self.extra = extra_args
        # init superclass
        StagedTaskCollection.__init__(self, name)


    def stage0(self):
        """Run the TURBOMOLE application specified to the constructor."""
        return self.turbomole_application


    def stage1(self):
        """Run 'xmllint'."""
        # terminate if first stage was unsuccessful
        rc = self.tasks[0].execution.returncode
        if rc is not None and rc != 0:
            return rc
        self.turbomole_output_dir = self.tasks[0].output_dir
        return XmlLintApplication(self.turbomole_output_dir,
                                  os.path.join(self.output_dir, 'xmllint'),
                                  **self.extra)


    def stage2(self):
        """Run 'eXist/client.sh'."""
        # terminate if first stage was unsuccessful
        rc = self.tasks[1].execution.returncode
        if rc is not None and rc != 0:
            return rc
        return XmlDbApplication(self.turbomole_output_dir,
                                os.path.join(self.output_dir, 'eXist'),
                                self.db_dir, self.db_user, self.db_pass,
                                **self.extra)


class BasisSweepPasses(StagedTaskCollection):
    """
    Build a two-step sequence:
      - first task is RIDFT with given coordinates and ``define.in`` file;
      - second task is a parallel collection of RICC2 that uses the output files from
        the first stage as input, plus a new ``define.in``.
    """
    def __init__(self, name, coord, ridft_in, ricc2_ins, work_dir, **extra_args):
        """
        Construct a new `BasisSweepPasses` sequential collection.

        :param str name: A string uniquely identifying this
        computation, to be used as a title in the TURBOMOLE input
        file.

        :param str coord: Path to the input ``coord`` file.

        :param str ridft_in: Path to the ``define.in`` file for the
        RIDFT step.

        :param ricc2_ins: Iterable, yielding the paths to the
        ``define.in`` file for each of the dependent RICC2 steps.

        :param str work_dir: Path to a directory where input files and
        results will be stored.

        """
        # need to remove this, we override it both in pass1 and pass2
        if extra_args.has_key('output_dir'):
            del extra_args['output_dir']
        # remember for later
        self.orb_basis = ridft_in._keywords['ORB_BASIS']
        self.rijk_basis = ridft_in._keywords['RIJK_BASIS']
        self.work_dir = os.path.join(work_dir, 'bas-%s/jkbas-%s'
                                     % (self.orb_basis, self.rijk_basis))
        self.name = name
        self.coord = coord
        self.ridft_in = ridft_in
        self.ricc2_ins = ricc2_ins
        self.extra = extra_args
        # init superclass
        StagedTaskCollection.__init__(self, name)


    def stage0(self):
        """
        Run a RIDFT job for the BAS/JKBAS combination given by the
        keywords ``ORB_BASIS`` and ``RIJK_BASIS`` in
        `self.ridft_in`.

        """
        # run 1st pass in the `ridft` directory
        gc3libs.utils.mkdir(self.work_dir)
        ridft_dir = os.path.join(self.work_dir, 'ridft')
        gc3libs.utils.mkdir(ridft_dir)
        gc3libs.utils.copyfile(self.coord, os.path.join(ridft_dir, 'coord'))
        ridft_define_in = _make_define_in(ridft_dir, self.ridft_in)
        ridft_output_dir =  os.path.join(ridft_dir, 'output')
        # application to run in pass 1
        gc3libs.log.debug("Creating RIDFT task '%s' (bas=%s, jkbas=%s) in directory '%s'",
                          self.name, self.orb_basis, self.rijk_basis, ridft_dir)
        # RIDFT expected to complete in 1 hour regardless
        extra = self.extra.copy()
        extra.setdefault('requested_walltime', 1*hours)
        return TurbomoleAndXmlProcessingPass(
            # job name
            ('ridft-%s-%s-%s' % (self.name, self.orb_basis, self.rijk_basis)),
            # TURBOMOLE application to run
            NonLocalTurbomoleDefineApplication(
                'ridft', ridft_define_in, self.coord,
                output_dir = ridft_output_dir,
                stdout = 'ridft.out',
                **extra),
            # base output directory for xmllint and eXist jobs
            ridft_dir,
            # DB parameters
            # FIXME: make these settable on the command-line
            db_dir='/db/home/fox/gricomp', db_user='fox', db_pass='tueR!?05',
            # TaskCollection required params
            **self.extra)


    def stage1(self):
        """
        Run a RICC2 job for each valid CBAS/CABS basis combination,
        re-using the results from RIDFT in `stage0`.

        If RIDFT failed, exit immediately.
        """
        # terminate if first stage was unsuccessful
        rc = self.tasks[0].execution.returncode
        if rc is not None and rc != 0:
            return rc
        # else, proceeed with 2nd pass
        pass2 = [ ]
        ridft_coord = os.path.join(self.tasks[0].turbomole_output_dir, 'coord')
        for ricc2_in in self.ricc2_ins:
            cbas = ricc2_in._keywords['CBAS_BASIS']
            cabs = ricc2_in._keywords['CABS_BASIS']
            ricc2_dir = os.path.join(self.work_dir,
                                     'cbas-%s/cabs-%s/ricc2' % (cbas, cabs))
            gc3libs.utils.mkdir(ricc2_dir)
            gc3libs.utils.copyfile(ridft_coord, ricc2_dir)
            ricc2_define_in = _make_define_in(ricc2_dir, ricc2_in)
            ricc2_output_dir = os.path.join(ricc2_dir, 'output')
            # guess duration of the RICC2 job
            extra = self.extra.copy()
            if ('aug-cc-pV5Z' == self.orb_basis
                or 'aug-cc-pV5Z' == self.rijk_basis
                or 'aug-cc-pV5Z' == cbas
                or 'aug-cc-pV5Z' == cabs):
                extra.setdefault('requested_walltime', 4*hours)
            else:
                extra.setdefault('requested_walltime', 1*hours)
            pass2.append(
                TurbomoleAndXmlProcessingPass(
                    # job name
                    ('ricc2-%s-%s-%s' % (self.name, cbas, cabs)),
                    # TURBOMOLE application to run
                    NonLocalTurbomoleDefineApplication(
                        'ricc2', ricc2_define_in,
                        # the second pass builds on files defined in the first one
                        os.path.join(ricc2_dir, 'coord'),
                        os.path.join(self.tasks[0].turbomole_output_dir, 'control'),
                        os.path.join(self.tasks[0].turbomole_output_dir, 'energy'),
                        os.path.join(self.tasks[0].turbomole_output_dir, 'mos'),
                        os.path.join(self.tasks[0].turbomole_output_dir, 'basis'),
                        os.path.join(self.tasks[0].turbomole_output_dir, 'auxbasis'),
                        output_dir = ricc2_output_dir,
                        stdout = 'ricc2.out',
                        **extra),
                    os.path.join(ricc2_output_dir, 'xml-processing'),
                    # DB parameters
                    # FIXME: make these settable on the command-line
                    db_dir='/db/home/fox/gricomp', db_user='fox', db_pass='tueR!?05',
                    # TaskCollection required params
                    **self.extra))
            gc3libs.log.debug("Created RICC2 task in directory '%s'", ricc2_dir)
        return (ParallelTaskCollection(self.name + '.pass2', pass2))


class BasisSweep(ParallelTaskCollection):
    """
    For each valid combination of bases, perform a RIDFT+RICC2
    analysis of the molecule given in `coord`.

    :param str title: A string to name this TURBOMOLE computation.
        Written unchanged into TURBOMOLE's ``define.in`` file, so it
        should only contain ASCII characters excluding control
        characters.

    :param str coord: Path to a file containing molecular coordinates
        in Turbomole format.

    :param list bases: Names of the orbital bases to sweep.

    :param list jkbases: Names of the RIJK bases to sweep.

    :param list cbases: Values for TURBOMOLE's `cbas` parameter to sweep.

    :param list cabses: Values for TURBOMOLE's `cabs` parameter to sweep.

    :param str work_dir: Path to a directory where input files and
        results will be stored.

    :param func valid1: A function taking a pair (orbital basis, jk
        basis) and returning `True` iff that combination is valid and
        should be analyzed.

    :param func valid2: A function taking a pair (cbas, cabs) and
        returning `True` iff that combination is valid and should be
        analyzed.
    """

    def __init__(self, title, coord, bases, jkbases, cbases, cabses, work_dir,
                 valid1=acceptable_ridft_basis_set,
                 valid2=acceptable_ricc2_basis_set,
                 **extra_args):
        """
        Create a new tasks that runs several analyses in parallel, one
        for each accepted combination of orbital and RIJK basis.
        """
        extra_args.setdefault('memory', 2000) # XXX: check with `requested_memory`

        ridft_define_in = Template(
            RIDFT_DEFINE_IN, valid1,
            TITLE=title,
            ORB_BASIS=bases,
            RIJK_BASIS=jkbases,
            RIDFT_MEMORY = [extra_args['memory']]
            ) # end of RIDFT template

        ricc2_define_in = Template(
            RICC2_DEFINE_IN, valid2,
            # the ORB_BASIS will be derived from the RIDFT_DEFINE_IN template
            CBAS_BASIS=cbases,
            CABS_BASIS=cabses,
            RICC2_MEMORY = [extra_args['memory']],
            ) # end of RICC2 template

        tasks = [ ]
        for ridft in expansions(ridft_define_in):
            orb_basis = ridft._keywords['ORB_BASIS']
            tasks.append(
                BasisSweepPasses(
                    title + '.seq', coord, ridft,
                    list(expansions(ricc2_define_in,
                                    ORB_BASIS=orb_basis)),
                    work_dir, **extra_args))
        ParallelTaskCollection.__init__(self, title, tasks)


## main

class GRICompScript(SessionBasedScript):
    """
For each molecule defined in a ``coord`` file given on the
command-line, run TURBOMOLE's ``ridft`` and then ``ricc2``
programs, with each possible combination of orbital and
auxiliary basis sets.

The list of orbital and auxiliary basis sets to try can be
controlled with the ``--bas``, ``--jkbas``, ``--cbas`` and
``--cabs`` options.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = '0.2',
            # TURBOMOLE's "coord" files are input
            input_filename_pattern = 'coord',
            )

    def setup_args(self):
        super(GRICompScript, self).setup_args()

        self.add_param("--bas", metavar="LIST", action="append",
                       dest="bas", default=[],
                       help="Comma-separated list of orbital bases to sweep."
                       " (Default: %(default)s")
        self.add_param("--jkbas", metavar="LIST", action="append",
                       dest="jkbas", default=['aug-cc-pVTZ', 'aug-cc-pVQZ', 'aug-cc-pV5Z'],
                       help="Comma-separated list of RIJK bases to sweep."
                       " (Default: %(default)s")
        self.add_param("--cbas", metavar="LIST", action="append",
                       dest="cbas", default=['aug-cc-pVTZ', 'aug-cc-pVQZ', 'aug-cc-pV5Z'],
                       help="Comma-separated list of `cbas` bases to sweep."
                       " (Default: %(default)s")
        self.add_param("--cabs", metavar="LIST", action="append",
                       dest="cabs", default=['aug-cc-pVTZ', 'aug-cc-pVQZ', 'aug-cc-pV5Z'],
                       help="Comma-separated list of `cabs` bases to sweep."
                       " (Default: %(default)s")


    def parse_args(self):
        # collect the basis set names given to the ``--bas``,
        # ``--jkbas``, ``--cbas`` and ``--cabs`` options and make them
        # into properly formatted lists.

        if len(self.params.bas) == 0:
            self.params.bas = basis_set_names
        else:
            self.params.bas = str.join(',', self.params.bas).split(',')
        for name in self.params.bas:
            if name not in basis_set_names:
                raise gc3libs.exceptions.InvalidUsage(
                    "Unknown basis set name: '%s'." % name)

        if len(self.params.jkbas) == 0:
            self.params.jkbas = jkbasis_set_names
        else:
            self.params.jkbas = str.join(',', self.params.jkbas).split(',')
        for name in self.params.jkbas:
            if name not in basis_set_names:
                raise gc3libs.exceptions.InvalidUsage(
                    "Unknown basis set name: '%s'." % name)

        if len(self.params.cbas) == 0:
            self.params.cbas = cbasis_set_names
        else:
            self.params.cbas = str.join(',', self.params.cbas).split(',')
        for name in self.params.cbas:
            if name not in basis_set_names:
                raise gc3libs.exceptions.InvalidUsage(
                    "Unknown basis set name: '%s'." % name)

        if len(self.params.cabs) == 0:
            self.params.cabs = cabsis_set_names
        else:
            self.params.cabs = str.join(',', self.params.cabs).split(',')
        for name in self.params.cabs:
            if name not in basis_set_names:
                raise gc3libs.exceptions.InvalidUsage(
                    "Unknown basis set name: '%s'." % name)


    def new_tasks(self, extra):
        coords = self._search_for_input_files(self.params.args)

        for coord in coords:
            # XXX: how do we get a unique name for each coord?  for
            # now, assume the directory containing the `coord` file
            # gives the unique name
            name = os.path.basename(os.path.dirname(coord))
            yield (name,
                   gricomp.BasisSweep, [
                       name,
                       coord,
                       self.params.bas,
                       self.params.jkbas,
                       self.params.cbas,
                       self.params.cabs,
                       self.make_directory_path(self.params.output, name),
                       ],
                   {})



#! /usr/bin/env python
#
"""
Support and expansion of programmatic templates.

The module `gc3libs.template` allows creation of textual templates
with a simple object-oriented programming interface: given a string
with a list of substitutions (using the syntax of Python's standard
`substitute` module), a set of replacements can be specified, and the
`gc3libs.template.expansions` function will generate all possible
texts coming from the same template.  Templates can be nested, and
expansions generated recursviely.
"""
# Copyright (C) 2009-2012, 2014 GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '2.2.1 version (SVN $Revision$)'


import string


try:
    import itertools
    SetProductIterator = itertools.product
except:
    # use our own implementation, in case `itertools` does not (yet)
    # provide the set-product
    class SetProductIterator(object):
        """Iterate over all elements in a cartesian product.

        Argument `factors` is a sequence, all whose items are sequences
        themselves: the returned iterator will return -upon each
        successive invocation- a list `[t_1, t_2, ..., t_n]` where `t_k`
        is an item in the `k`-th sequence.

        Examples::
          >>> list(SetProductIterator([]))
          [[]]
          >>> list(SetProductIterator([1]))
          [[1]]
          >>> list(SetProductIterator([1],[1]))
          [[1, 1]]
          >>> list(SetProductIterator([1,2],[]))
          [[]]
          >>> list(SetProductIterator([1,2],[1]))
          [[1, 1], [2, 1]]
          >>> list(SetProductIterator([1, 2], [1, 2]))
          [[1, 1], [2, 1], [1, 2], [2, 2]]
        """
        def __init__(self, *factors):
            self.__closed = False
            self.__factors = factors
            self.__L = len(factors)
            self.__M = [ len(s)-1 for s in factors ]
            self.__m = [0] * self.__L
            self.__i = 0

        def __iter__(self):
            return self

        def next(self):
            if self.__closed:
                raise StopIteration
            if (0 == self.__L) or (-1 in self.__M):
                # there are no factors, or one of them has no elements
                self.__closed = True
                return []
            else:
                if self.__i < self.__L:
                    # will return element corresponding to current multi-index
                    result = [ s[self.__m[i]]
                               for (i,s) in enumerate(self.__factors) ]
                    # advance multi-index
                    i = 0
                    while (i < self.__L):
                        if self.__m[i] == self.__M[i]:
                            self.__m[i] = 0
                            i += 1
                        else:
                            self.__m[i] += 1
                            break
                    self.__i = i
                    # back to caller
                    return result
                else:
                    # at end of iteration
                    self.__closed = True
                    raise StopIteration


class Template(object):
    """
    A template object is a pair `(obj, keywords)`.  Methods are
    provided to substitute the keyword values into `obj`, and to
    iterate over expansions of the given keywords (optionally
    filtering the allowed combination of keyword values).

    Second optional argument `validator` must be a function that
    accepts a set of keyword arguments, and returns `True` if the
    keyword combination is valid (can be expanded/substituted back
    into the template) or `False` if it should be discarded.
    The default validator passes any combination of keywords/values.
    """
    def __init__(self, template,
                 validator=(lambda kws: True), **extra_args):
        self._template = template
        self._keywords = extra_args
        self._valid = validator

    def substitute(self, **extra_args):
        """
        Return result of interpolating the value of keywords into the
        template.  Keyword arguments `extra_args` can be used to override
        keyword values passed to the constructor.

        If the templated object provides a `substitute` method, then
        return the result of invoking it with the template keywords as
        keyword arguments.  Otherwise, return the result of applying
        Python standard library's `string.Template.safe_substitute()`
        on the string representation of the templated object.

        Raise `ValueError` if the set of keywords/values is not valid
        according to the validator specified in the constructor.
        """
        keywords = self._keywords.copy()
        keywords.update(extra_args)
        if self._valid(keywords):
            try:
                return self._template.substitute(**keywords)
            except AttributeError:
                return string.Template(str(self._template)).safe_substitute(keywords)
        else:
            raise ValueError("Invalid substitution values in template.")

    def __str__(self):
        """Alias for `Template.substitute`."""
        return self.substitute()

    def __repr__(self):
        """
        Return a string representation such that `x == eval(repr(x))`.
        """
        return str.join('', ["Template(",
                             str.join(', ', [repr(self._template)] +
                                      [ ("%s=%s" % (k,v))
                                        for k,v in self._keywords.items() ]),
                             ')'])


    def expansions(self, **keywords):
        """
        Iterate over all valid expansions of the templated object
        *and* the template keywords.  Returned items are `Template`
        instances constucted with the expanded template object and a
        valid combination of keyword values.
        """
        keywords_ = self._keywords.copy()
        keywords_.update(keywords)
        for kws in expansions(keywords_):
            for item in expansions(self._template, **kws):
                # propagate expanded keywords upwards;
                # this allows container templates to
                # check validity based on keywords expanded
                # on contained templates as well.
                new_kws = kws.copy()
                for v in kws.values():
                    if isinstance(v, Template):
                        new_kws.update(v._keywords)
                if self._valid(new_kws):
                    yield Template(item, self._valid, **new_kws)


def expansions(obj, **extra_args):
    """
    Iterate over all expansions of a given object, recursively
    expanding all templates found.  How the expansions are actually
    computed, depends on the type of object being passed in the first
    argument `obj`:

    * If `obj` is a `list`, iterate over expansions of items in `obj`.
      (In particular, this flattens out nested lists.)

      Example::

        >>> L = [0, [2, 3]]
        >>> list(expansions(L))
        [0, 2, 3]

    * If `obj` is a dictionary, return dictionary formed by all
      combinations of a key `k` in `obj` with an expansion of the
      corresponding value `obj[k]`.  Expansions are computed by
      recursively calling `expansions(obj[k], **extra_args)`.

      Example::

        >>> D = {'a':1, 'b':[2,3]}
        >>> E = list(expansions(D))
        >>> len(E)
        2
        >>> {'a': 1, 'b': 2} in E
        True
        >>> {'a': 1, 'b': 3} in E
        True

    * If `obj` is a `tuple`, iterate over all tuples formed by the
      expansion of every item in `obj`.  (Each item `t[i]` is expanded
      by calling `expansions(t[i], **extra_args)`.)

      Example::

        >>> T = (1, [2, 3])
        >>> list(expansions(T))
        [(1, 2), (1, 3)]

    * If `obj` is a `Template` class instance, then the returned values
      are the result of applying the template to the expansion of each
      of its keywords.

      Example::

        >>> T1 = Template("a=${n}", n=[0,1])
        >>> list(expansions(T1))
        [Template('a=${n}', n=0), Template('a=${n}', n=1)]

      Note that keywords passed to the `expand` invocation override
      the ones used in template construction::

        >>> T2 = Template("a=${n}")
        >>> list(expansions(T2, n=[1,3]))
        [Template('a=${n}', n=1), Template('a=${n}', n=3)]

        >>> T3 = Template("a=${n}", n=[0,1])
        >>> list(expansions(T3, n=[2,3]))
        [Template('a=${n}', n=2), Template('a=${n}', n=3)]

    * Any other value is returned unchanged.

      Example:

        >>> V = 42
        >>> list(expansions(V))
        [42]

    """
    if isinstance(obj, dict):
        keys = tuple(obj.keys()) # fix a key order
        for items in SetProductIterator(*[ list(expansions(obj[keys[i]], **extra_args))
                                           for i in range(len(keys)) ]):
            yield dict((keys[i], items[i]) for i in range(len(keys)))
    elif isinstance(obj, tuple):
        for items in SetProductIterator(*[ list(expansions(u, **extra_args)) for u in obj ]):
            yield tuple(items)
    elif isinstance(obj, list):
        for item in obj:
            for result in expansions(item, **extra_args):
                yield result
    elif isinstance(obj, Template):
        for s in obj.expansions(**extra_args):
            yield s
    else:
        yield obj



## main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="template",
                    optionflags=doctest.NORMALIZE_WHITESPACE)
